import os
import uuid
from typing import Any


def _organization_directory_path(org: Any, filename: str) -> str:
    """
    Create a directory path to upload the organization logo

    """

    org_name = org.name
    _, extension = os.path.splitext(filename)
    return f"files/public/organizations/{org_name}/{extension[1:]}/{filename}"


def _chart_image_directory_path(chart: Any, filename: str) -> str:
    """
    Create a directory path to upload the organization logo

    """

    chart_name = chart.name
    _, extension = os.path.splitext(filename)
    return f"files/chart_images/{chart_name}/{extension[1:]}/{filename}"


def _dataspace_directory_path(ds: Any, filename: str) -> str:
    """
    Create a directory path to upload the dataspace logo

    """

    ds_name = ds.name
    _, extension = os.path.splitext(filename)
    return f"files/public/dataspace/{ds_name}/{extension[1:]}/{filename}"


def _use_case_directory_path(uc: Any, filename: str) -> str:
    """
    Create a directory path to upload the use case logo

    """

    uc_name = uc.title
    _, extension = os.path.splitext(filename)
    return f"files/use_case/{uc_name}/logo/{filename}"


def _user_profile_image_directory_path(user: Any, filename: str) -> str:
    """
    Create a directory path to upload the user profile image

    """
    user_name = user.keycloak_id
    _, extension = os.path.splitext(filename)
    return f"files/user_profile/{user_name}/{extension[1:]}/{filename}"


def _organization_file_directory_path(org: Any, filename: str) -> str:
    """
    Create a directory path to upload the sample data file.

    """

    org_name = org.name
    _, extension = os.path.splitext(filename)
    return f"files/resources/{org_name}/sample_data/{extension[1:]}/{filename}"


def _catalog_directory_path(catalog: Any, filename: str) -> str:
    """
    Create a directory path to upload the catalog logo
    """
    catalog_name = catalog.name
    _, extension = os.path.splitext(filename)
    return f"files/catalog/{catalog_name}/{extension[1:]}/{filename}"



"""Type stubs for django-cors-headers.middleware"""

from typing import Any, Callable, List, Optional, Set, Union

from django.http import HttpRequest, HttpResponse

class CorsMiddleware:
    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]) -> None: ...
    def __call__(self, request: HttpRequest) -> HttpResponse: ...
    def process_response(
        self, request: HttpRequest, response: HttpResponse
    ) -> HttpResponse: ...



import base64
import json
import pickle
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    cast,
)

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models

JSONValue = Union[Dict[str, Any], List[Any], str, int, float, bool, None]
T = TypeVar("T")


class CustomJSONEncoder(DjangoJSONEncoder):
    """Custom JSON encoder that handles non-standard objects using pickle."""

    def default(self, obj: Any) -> JSONValue:
        """Only serialize non-standard objects using pickle."""
        if isinstance(obj, (dict, list, str, int, float, bool, type(None))):
            return obj  # JSON-compatible values are returned as-is

        # For non-standard objects, serialize using pickle and base64 encoding
        return {
            "_custom_serialized": True,
            "data": base64.b64encode(pickle.dumps(obj)).decode("utf-8"),
        }


class CustomJSONDecoder:
    """Custom JSON decoder that handles pickled objects."""

    @staticmethod
    def decode(value: Any) -> Any:
        """Decode objects that were custom-serialized, otherwise return normal JSON."""
        if isinstance(value, dict) and value.get("_custom_serialized"):
            try:
                return pickle.loads(base64.b64decode(value["data"].encode("utf-8")))
            except (pickle.PickleError, ValueError):
                return value  # Fallback to returning original if decoding fails
        return value


class SerializableJSONField(models.JSONField):
    """JSON field that can handle non-standard Python objects."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.default: Callable[[], Union[Dict[str, Any], List[Any]]] = kwargs.get(
            "default", dict
        )

    def from_db_value(
        self, value: Optional[str], expression: Any, connection: Any
    ) -> Union[Dict[str, Any], List[Any]]:
        """Ensure that deserialized values match their expected types"""
        if value is None:
            return []
        if isinstance(value, str):
            try:
                data = json.loads(value)  # Convert from string to JSON
                decoded_data = self._decode_values(data)
                return decoded_data if isinstance(decoded_data, (list, dict)) else []
            except (json.JSONDecodeError, TypeError):
                return []  # Return empty list if decoding fails
        return cast(Union[Dict[str, Any], List[Any]], value)

    def get_prep_value(self, value: Any) -> Optional[str]:
        """Automatically serialize before saving to database"""
        if value is None:
            return value
        if isinstance(value, str):  # Prevent double encoding
            return value
        return json.dumps(self._encode_values(value), cls=CustomJSONEncoder)

    def _encode_values(self, obj: Any) -> Any:
        """Ensure that only non-JSON serializable objects are encoded"""
        if isinstance(obj, (dict, list, str, int, float, bool, type(None))):
            return obj  # Return JSON-native types as-is
        return CustomJSONEncoder().default(obj)  # Encode non-serializable objects

    def _decode_values(self, obj: Any) -> Any:
        """Ensure proper decoding of serialized objects and remove _custom_serialized fields"""
        if isinstance(obj, dict):
            if "_custom_serialized" in obj and "data" in obj:
                # Deserialize the object and return the decoded value
                return CustomJSONDecoder.decode(obj)

            # Recursively decode and remove _custom_serialized fields from all dictionary items
            return {
                key: self._decode_values(value)
                for key, value in obj.items()
                if key != "_custom_serialized"
            }

        if isinstance(obj, list):
            return [self._decode_values(value) for value in obj]

        if isinstance(obj, str):
            try:
                return self._decode_values(
                    json.loads(obj)
                )  # If it's a JSON string, load it as a Python object
            except json.JSONDecodeError:
                return obj  # Otherwise, return as-is

        return obj  # Return original object if it doesn't match any criteria

    def __getitem__(self, key: str) -> Any:
        """Make the field indexable."""
        if not hasattr(self, "value"):
            raise KeyError(key)
        value = cast(Dict[str, Any], self.value)
        return value[key]

    def get(self, key: str, default: Optional[T] = None) -> Optional[T]:
        """Get a value from the field by key."""
        if not hasattr(self, "value"):
            return default
        value = cast(Dict[str, Any], self.value)
        return value.get(key, default)  # type: ignore



from api.types.charts.big_number_chart import BigNumberChart
from api.types.charts.enhanced_map_chart import (
    GeospatialMapChart,
    PointMapChart,
    PolygonMapChart,
)
from api.types.charts.map_chart import MapChart
from api.types.charts.treemap_chart import TreeMapChart
from api.types.charts.unified_chart import UnifiedChart



from typing import Optional

import strawberry
import strawberry_django
from strawberry import auto

from api.models import ResourceChartImage
from api.types.base_type import BaseType
from api.types.type_dataset import TypeDataset
from api.utils.enums import ChartStatus


@strawberry_django.filter(ResourceChartImage)
class ResourceChartImageFilter:
    id: auto
    name: auto


ChartStatusEnum = strawberry.enum(ChartStatus)  # type: ignore


@strawberry_django.type(
    ResourceChartImage,
    pagination=True,
    fields="__all__",
    filters=ResourceChartImageFilter,
)
class TypeResourceChartImage(BaseType):
    modified: auto
    description: Optional[str] = ""
    dataset: Optional[TypeDataset] = None
    status: ChartStatusEnum = ChartStatusEnum.DRAFT



"""
Django management command to populate Asia Pacific geography data.
Covers: India, Indonesia, Thailand, Philippines at state/province level.

Usage:
    python manage.py populate_geographies
    python manage.py populate_geographies --clear  # Clear existing data first
"""

from typing import Any

from django.core.management.base import BaseCommand, CommandParser
from django.db import transaction

from api.models import Geography
from api.utils.enums import GeoTypes


class Command(BaseCommand):
    help = (
        "Populate Asia Pacific geography data (India, Indonesia, Thailand, Philippines)"
    )

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "--clear",
            action="store_true",
            help="Clear all existing geography data before populating",
        )

    def handle(self, *args: Any, **options: Any) -> None:
        clear_existing = options.get("clear", False)

        if clear_existing:
            self.stdout.write(self.style.WARNING("Clearing existing geography data..."))
            Geography.objects.all().delete()
            self.stdout.write(self.style.SUCCESS("✓ Cleared existing data"))

        with transaction.atomic():
            self._populate_geographies()

        self.stdout.write(
            self.style.SUCCESS(
                f"\n✅ Successfully populated {Geography.objects.count()} geography entries"
            )
        )

    def _populate_geographies(self) -> None:
        """Populate geography data in an idempotent way."""

        # Create or get Asia Pacific region
        asia_pacific, created = Geography.objects.get_or_create(
            name="Asia Pacific", defaults={"code": "APAC", "type": GeoTypes.REGION}
        )
        if created:
            self.stdout.write(f"Created region: {asia_pacific.name}")
        else:
            self.stdout.write(f"Region already exists: {asia_pacific.name}")

        # India and its states
        self._populate_india(asia_pacific)

        # Indonesia and its provinces
        self._populate_indonesia(asia_pacific)

        # Thailand and its provinces
        self._populate_thailand(asia_pacific)

        # Philippines and its provinces
        self._populate_philippines(asia_pacific)

    def _populate_india(self, parent: Geography) -> None:
        """Populate India and its states/UTs."""
        india, created = Geography.objects.get_or_create(
            name="India",
            defaults={"code": "IN", "type": GeoTypes.COUNTRY, "parent_id": parent},
        )
        if created:
            self.stdout.write(f"Created country: {india.name}")
        else:
            self.stdout.write(f"Country already exists: {india.name}")

        indian_states = [
            ("Andhra Pradesh", "AP"),
            ("Arunachal Pradesh", "AR"),
            ("Assam", "AS"),
            ("Bihar", "BR"),
            ("Chhattisgarh", "CT"),
            ("Goa", "GA"),
            ("Gujarat", "GJ"),
            ("Haryana", "HR"),
            ("Himachal Pradesh", "HP"),
            ("Jharkhand", "JH"),
            ("Karnataka", "KA"),
            ("Kerala", "KL"),
            ("Madhya Pradesh", "MP"),
            ("Maharashtra", "MH"),
            ("Manipur", "MN"),
            ("Meghalaya", "ML"),
            ("Mizoram", "MZ"),
            ("Nagaland", "NL"),
            ("Odisha", "OR"),
            ("Punjab", "PB"),
            ("Rajasthan", "RJ"),
            ("Sikkim", "SK"),
            ("Tamil Nadu", "TN"),
            ("Telangana", "TG"),
            ("Tripura", "TR"),
            ("Uttar Pradesh", "UP"),
            ("Uttarakhand", "UT"),
            ("West Bengal", "WB"),
            ("Andaman and Nicobar Islands", "AN"),
            ("Chandigarh", "CH"),
            ("Dadra and Nagar Haveli and Daman and Diu", "DN"),
            ("Delhi", "DL"),
            ("Jammu and Kashmir", "JK"),
            ("Ladakh", "LA"),
            ("Lakshadweep", "LD"),
            ("Puducherry", "PY"),
        ]

        created_count = 0
        for state_name, state_code in indian_states:
            _, created = Geography.objects.get_or_create(
                name=state_name,
                defaults={
                    "code": state_code,
                    "type": GeoTypes.STATE,
                    "parent_id": india,
                },
            )
            if created:
                created_count += 1

        self.stdout.write(
            f"Indian states/UTs: {created_count} created, {len(indian_states) - created_count} already existed"
        )

    def _populate_indonesia(self, parent: Geography) -> None:
        """Populate Indonesia and its provinces."""
        indonesia, created = Geography.objects.get_or_create(
            name="Indonesia",
            defaults={"code": "ID", "type": GeoTypes.COUNTRY, "parent_id": parent},
        )
        if created:
            self.stdout.write(f"Created country: {indonesia.name}")
        else:
            self.stdout.write(f"Country already exists: {indonesia.name}")

        indonesian_provinces = [
            ("Aceh", "AC"),
            ("Bali", "BA"),
            ("Banten", "BT"),
            ("Bengkulu", "BE"),
            ("Central Java", "JT"),
            ("Central Kalimantan", "KT"),
            ("Central Sulawesi", "ST"),
            ("East Java", "JI"),
            ("East Kalimantan", "KI"),
            ("East Nusa Tenggara", "NT"),
            ("Gorontalo", "GO"),
            ("Jakarta", "JK"),
            ("Jambi", "JA"),
            ("Lampung", "LA"),
            ("Maluku", "MA"),
            ("North Kalimantan", "KU"),
            ("North Maluku", "MU"),
            ("North Sulawesi", "SA"),
            ("North Sumatra", "SU"),
            ("Papua", "PA"),
            ("Riau", "RI"),
            ("Riau Islands", "KR"),
            ("South Kalimantan", "KS"),
            ("South Sulawesi", "SN"),
            ("South Sumatra", "SS"),
            ("Southeast Sulawesi", "SG"),
            ("West Java", "JB"),
            ("West Kalimantan", "KB"),
            ("West Nusa Tenggara", "NB"),
            ("West Papua", "PB"),
            ("West Sulawesi", "SR"),
            ("West Sumatra", "SB"),
            ("Yogyakarta", "YO"),
        ]

        created_count = 0
        for province_name, province_code in indonesian_provinces:
            _, created = Geography.objects.get_or_create(
                name=province_name,
                defaults={
                    "code": province_code,
                    "type": GeoTypes.STATE,
                    "parent_id": indonesia,
                },
            )
            if created:
                created_count += 1

        self.stdout.write(
            f"Indonesian provinces: {created_count} created, {len(indonesian_provinces) - created_count} already existed"
        )

    def _populate_thailand(self, parent: Geography) -> None:
        """Populate Thailand and its provinces."""
        thailand, created = Geography.objects.get_or_create(
            name="Thailand",
            defaults={"code": "TH", "type": GeoTypes.COUNTRY, "parent_id": parent},
        )
        if created:
            self.stdout.write(f"Created country: {thailand.name}")
        else:
            self.stdout.write(f"Country already exists: {thailand.name}")

        thai_provinces = [
            ("Bangkok", "BKK"),
            ("Amnat Charoen", "37"),
            ("Ang Thong", "15"),
            ("Bueng Kan", "38"),
            ("Buriram", "31"),
            ("Chachoengsao", "24"),
            ("Chai Nat", "18"),
            ("Chaiyaphum", "36"),
            ("Chanthaburi", "22"),
            ("Chiang Mai", "50"),
            ("Chiang Rai", "57"),
            ("Chonburi", "20"),
            ("Chumphon", "86"),
            ("Kalasin", "46"),
            ("Kamphaeng Phet", "62"),
            ("Kanchanaburi", "71"),
            ("Khon Kaen", "40"),
            ("Krabi", "81"),
            ("Lampang", "52"),
            ("Lamphun", "51"),
            ("Loei", "42"),
            ("Lopburi", "16"),
            ("Mae Hong Son", "58"),
            ("Maha Sarakham", "44"),
            ("Mukdahan", "49"),
            ("Nakhon Nayok", "26"),
            ("Nakhon Pathom", "73"),
            ("Nakhon Phanom", "48"),
            ("Nakhon Ratchasima", "30"),
            ("Nakhon Sawan", "60"),
            ("Nakhon Si Thammarat", "80"),
            ("Nan", "55"),
            ("Narathiwat", "96"),
            ("Nong Bua Lamphu", "39"),
            ("Nong Khai", "43"),
            ("Nonthaburi", "12"),
            ("Pathum Thani", "13"),
            ("Pattani", "94"),
            ("Phang Nga", "82"),
            ("Phatthalung", "93"),
            ("Phayao", "56"),
            ("Phetchabun", "67"),
            ("Phetchaburi", "76"),
            ("Phichit", "66"),
            ("Phitsanulok", "65"),
            ("Phra Nakhon Si Ayutthaya", "14"),
            ("Phrae", "54"),
            ("Phuket", "83"),
            ("Prachinburi", "25"),
            ("Prachuap Khiri Khan", "77"),
            ("Ranong", "85"),
            ("Ratchaburi", "70"),
            ("Rayong", "21"),
            ("Roi Et", "45"),
            ("Sa Kaeo", "27"),
            ("Sakon Nakhon", "47"),
            ("Samut Prakan", "11"),
            ("Samut Sakhon", "74"),
            ("Samut Songkhram", "75"),
            ("Saraburi", "19"),
            ("Satun", "91"),
            ("Sing Buri", "17"),
            ("Sisaket", "33"),
            ("Songkhla", "90"),
            ("Sukhothai", "64"),
            ("Suphan Buri", "72"),
            ("Surat Thani", "84"),
            ("Surin", "32"),
            ("Tak", "63"),
            ("Trang", "92"),
            ("Trat", "23"),
            ("Ubon Ratchathani", "34"),
            ("Udon Thani", "41"),
            ("Uthai Thani", "61"),
            ("Uttaradit", "53"),
            ("Yala", "95"),
            ("Yasothon", "35"),
        ]

        created_count = 0
        for province_name, province_code in thai_provinces:
            _, created = Geography.objects.get_or_create(
                name=province_name,
                defaults={
                    "code": province_code,
                    "type": GeoTypes.STATE,
                    "parent_id": thailand,
                },
            )
            if created:
                created_count += 1

        self.stdout.write(
            f"Thai provinces: {created_count} created, {len(thai_provinces) - created_count} already existed"
        )

    def _populate_philippines(self, parent: Geography) -> None:
        """Populate Philippines and its provinces."""
        philippines, created = Geography.objects.get_or_create(
            name="Philippines",
            defaults={"code": "PH", "type": GeoTypes.COUNTRY, "parent_id": parent},
        )
        if created:
            self.stdout.write(f"Created country: {philippines.name}")
        else:
            self.stdout.write(f"Country already exists: {philippines.name}")

        philippine_provinces = [
            # Luzon - NCR
            ("Metro Manila", "NCR"),
            # Luzon - CAR
            ("Abra", "ABR"),
            ("Apayao", "APA"),
            ("Benguet", "BEN"),
            ("Ifugao", "IFU"),
            ("Kalinga", "KAL"),
            ("Mountain Province", "MOU"),
            # Luzon - Region I (Ilocos)
            ("Ilocos Norte", "ILN"),
            ("Ilocos Sur", "ILS"),
            ("La Union", "LUN"),
            ("Pangasinan", "PAN"),
            # Luzon - Region II (Cagayan Valley)
            ("Batanes", "BTN"),
            ("Cagayan", "CAG"),
            ("Isabela", "ISA"),
            ("Nueva Vizcaya", "NUV"),
            ("Quirino", "QUI"),
            # Luzon - Region III (Central Luzon)
            ("Aurora", "AUR"),
            ("Bataan", "BAN"),
            ("Bulacan", "BUL"),
            ("Nueva Ecija", "NUE"),
            ("Pampanga", "PAM"),
            ("Tarlac", "TAR"),
            ("Zambales", "ZMB"),
            # Luzon - Region IV-A (CALABARZON)
            ("Batangas", "BTG"),
            ("Cavite", "CAV"),
            ("Laguna", "LAG"),
            ("Quezon", "QUE"),
            ("Rizal", "RIZ"),
            # Luzon - Region IV-B (MIMAROPA)
            ("Marinduque", "MAD"),
            ("Occidental Mindoro", "MDC"),
            ("Oriental Mindoro", "MDR"),
            ("Palawan", "PLW"),
            ("Romblon", "ROM"),
            # Luzon - Region V (Bicol)
            ("Albay", "ALB"),
            ("Camarines Norte", "CAN"),
            ("Camarines Sur", "CAS"),
            ("Catanduanes", "CAT"),
            ("Masbate", "MAS"),
            ("Sorsogon", "SOR"),
            # Visayas - Region VI (Western Visayas)
            ("Aklan", "AKL"),
            ("Antique", "ANT"),
            ("Capiz", "CAP"),
            ("Guimaras", "GUI"),
            ("Iloilo", "ILI"),
            ("Negros Occidental", "NEC"),
            # Visayas - Region VII (Central Visayas)
            ("Bohol", "BOH"),
            ("Cebu", "CEB"),
            ("Negros Oriental", "NER"),
            ("Siquijor", "SIG"),
            # Visayas - Region VIII (Eastern Visayas)
            ("Biliran", "BIL"),
            ("Eastern Samar", "EAS"),
            ("Leyte", "LEY"),
            ("Northern Samar", "NSA"),
            ("Samar", "WSA"),
            ("Southern Leyte", "SLE"),
            # Mindanao - Region IX (Zamboanga Peninsula)
            ("Zamboanga del Norte", "ZAN"),
            ("Zamboanga del Sur", "ZAS"),
            ("Zamboanga Sibugay", "ZSI"),
            # Mindanao - Region X (Northern Mindanao)
            ("Bukidnon", "BUK"),
            ("Camiguin", "CAM"),
            ("Lanao del Norte", "LAN"),
            ("Misamis Occidental", "MSC"),
            ("Misamis Oriental", "MSR"),
            # Mindanao - Region XI (Davao)
            ("Davao de Oro", "COM"),
            ("Davao del Norte", "DAV"),
            ("Davao del Sur", "DAS"),
            ("Davao Occidental", "DAO"),
            ("Davao Oriental", "DAO"),
            # Mindanao - Region XII (SOCCSKSARGEN)
            ("Cotabato", "NCO"),
            ("Sarangani", "SAR"),
            ("South Cotabato", "SCO"),
            ("Sultan Kudarat", "SUK"),
            # Mindanao - Region XIII (Caraga)
            ("Agusan del Norte", "AGN"),
            ("Agusan del Sur", "AGS"),
            ("Dinagat Islands", "DIN"),
            ("Surigao del Norte", "SUN"),
            ("Surigao del Sur", "SUR"),
            # Mindanao - BARMM
            ("Basilan", "BAS"),
            ("Lanao del Sur", "LAS"),
            ("Maguindanao", "MAG"),
            ("Sulu", "SLU"),
            ("Tawi-Tawi", "TAW"),
        ]

        created_count = 0
        for province_name, province_code in philippine_provinces:
            _, created = Geography.objects.get_or_create(
                name=province_name,
                defaults={
                    "code": province_code,
                    "type": GeoTypes.STATE,
                    "parent_id": philippines,
                },
            )
            if created:
                created_count += 1

        self.stdout.write(
            f"Philippine provinces: {created_count} created, {len(philippine_provinces) - created_count} already existed"
        )



"""
Hugging Face Client for running local or remote model inference.
Supports both pipeline-based and model-class-based inference,
and integrates with Django model management (AIModel).
"""

import time
from typing import Any, Dict, Optional, Tuple

import torch  # type: ignore
from asgiref.sync import sync_to_async
from django.utils import timezone
from tenacity import retry, stop_after_attempt, wait_exponential  # type: ignore
from transformers import (  # type: ignore
    AutoModelForCausalLM,
    AutoModelForMaskedLM,
    AutoModelForMultipleChoice,
    AutoModelForNextSentencePrediction,
    AutoModelForQuestionAnswering,
    AutoModelForSeq2SeqLM,
    AutoModelForSequenceClassification,
    AutoModelForTokenClassification,
    AutoTokenizer,
    pipeline,
)

from api.models import AIModel


class ModelHFClient:
    """Client for interacting with Hugging Face models."""

    def __init__(self, model: AIModel):
        self.model = model
        self.device = self._get_device()
        self.model_map = {
            "AutoModelForCausalLM": AutoModelForCausalLM,
            "AutoModelForSeq2SeqLM": AutoModelForSeq2SeqLM,
            "AutoModelForSequenceClassification": AutoModelForSequenceClassification,
            "AutoModelForNextSentencePrediction": AutoModelForNextSentencePrediction,
            "AutoModelForMultipleChoice": AutoModelForMultipleChoice,
            "AutoModelForTokenClassification": AutoModelForTokenClassification,
            "AutoModelForQuestionAnswering": AutoModelForQuestionAnswering,
            "AutoModelForMaskedLM": AutoModelForMaskedLM,
        }
        self.task_map = {
            "TRANSLATION": "translation",
            "TEXT_GENERATION": "text-generation",
            "SUMMARIZATION": "summarization",
            "QUESTION_ANSWERING": "question-answering",
            "SENTIMENT_ANALYSIS": "sentiment-analysis",
            "TEXT_CLASSIFICATION": "text-classification",
            "NAMED_ENTITY_RECOGNITION": "ner",
            "TEXT_TO_SPEECH": "text-to-speech",
            "SPEECH_TO_TEXT": "automatic-speech-recognition",
            "OTHER": "",
        }

    def _get_device(self) -> int:
        """Select device (0 for GPU if available, else CPU)."""
        return 0 if torch.cuda.is_available() else -1

    def _load_pipeline(self) -> Any:
        """Initialize a Hugging Face pipeline."""
        return pipeline(
            task=self.task_map.get(self.model.model_type, "text-generation"),
            model=self.model.provider_model_id,
            framework="pt",
            device=self.device,
            trust_remote_code=True,
            use_auth_token=self.model.hf_auth_token or None,
        )

    def _load_model_and_tokenizer(self) -> Tuple[Any, Any]:
        """Load model and tokenizer for manual inference."""
        tokenizer = AutoTokenizer.from_pretrained(
            self.model.provider_model_id,
            trust_remote_code=True,
            use_auth_token=self.model.hf_auth_token or None,
        )

        model_class = self.model_map.get(
            self.model.hf_model_class or "AutoModelForCausalLM", AutoModelForCausalLM
        )
        model = model_class.from_pretrained(
            pretrained_model_name_or_path=self.model.provider_model_id,
            trust_remote_code=True,
            torch_dtype=torch.float16 if torch.cuda.is_available() else torch.float32,
            use_auth_token=self.model.hf_auth_token or None,
            attn_implementation=self.model.hf_attn_implementation,
            device_map="auto",
        )

        return model, tokenizer

    def _generate_from_model(self, model: Any, tokenizer: Any, input_text: str) -> str:
        """Run generation with tokenizer/model."""
        inputs = tokenizer(
            input_text,
            padding=True,
            truncation=True,
            return_tensors="pt",
        )

        with torch.no_grad():
            outputs = model.generate(**inputs, use_cache=False)

        decoded_text: str = tokenizer.batch_decode(outputs, skip_special_tokens=True)[0]
        return decoded_text

    async def _update_success(self) -> None:
        """Update endpoint statistics for successful inference."""

        def _update() -> None:
            endpoint = self.model.get_primary_endpoint()
            if endpoint:
                endpoint.total_requests += 1
                endpoint.last_success_at = timezone.now()
                endpoint.save(update_fields=["total_requests", "last_success_at"])

        await sync_to_async(_update)()

    async def _update_failure(self) -> None:
        """Update endpoint statistics for failed inference."""

        def _update() -> None:
            endpoint = self.model.get_primary_endpoint()
            if endpoint:
                endpoint.failed_requests += 1
                endpoint.last_failure_at = timezone.now()
                endpoint.save(update_fields=["failed_requests", "last_failure_at"])

        await sync_to_async(_update)()

    @retry(stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=2, max=10))
    async def call_async(self, input_text: str) -> Dict[str, Any]:
        """
        Run asynchronous inference on the model.
        Supports both pipeline and manual model modes.
        """
        start_time = time.time()
        try:
            if self.model.hf_use_pipeline:
                pipe = self._load_pipeline()
                result = pipe(input_text)
                output_text = result if isinstance(result, str) else str(result)
            else:
                model, tokenizer = self._load_model_and_tokenizer()
                output_text = self._generate_from_model(model, tokenizer, input_text)

            latency_ms = (time.time() - start_time) * 1000
            await self._update_success()

            return {
                "success": True,
                "output": output_text,
                "latency_ms": latency_ms,
                "provider": "HuggingFace",
                "model_id": self.model.provider_model_id,
            }

        except Exception as e:
            await self._update_failure()
            return {
                "success": False,
                "error": str(e),
                "latency_ms": (time.time() - start_time) * 1000,
                "provider": "HuggingFace",
                "model_id": self.model.provider_model_id,
            }

    def call(self, input_text: str) -> Dict[str, Any]:
        """Run synchronous inference by wrapping async call."""
        import asyncio

        import nest_asyncio  # type: ignore

        try:
            # Case 1: We're in a normal sync context (Django shell, Python script)
            return asyncio.run(self.call_async(input_text))
        except RuntimeError as e:
            if "asyncio.run() cannot be called from a running event loop" in str(e):
                # Case 2: We're inside Jupyter/IPython or any async event loop
                nest_asyncio.apply()
                loop = asyncio.get_event_loop()
                return loop.run_until_complete(self.call_async(input_text))
            elif "no current event loop" in str(e).lower():
                # Case 3: No loop exists (Python ≥3.12)
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                return loop.run_until_complete(self.call_async(input_text))
            else:
                raise



import structlog
from rest_framework import status, views
from rest_framework.permissions import AllowAny, IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework_simplejwt.tokens import RefreshToken

from authorization.consent import UserConsent
from authorization.keycloak import keycloak_manager
from authorization.models import User
from authorization.serializers import UserConsentSerializer
from authorization.services import AuthorizationService

logger = structlog.getLogger(__name__)


class KeycloakLoginView(views.APIView):
    """
    View for handling Keycloak login and token exchange.
    Accepts Keycloak tokens and creates Django tokens.
    """

    permission_classes = [AllowAny]

    def post(self, request: Request) -> Response:

        # Get the Keycloak token from the request
        keycloak_token = request.data.get("token")
        if not keycloak_token:
            logger.warning("Login attempt without token")
            return Response(
                {"error": "Keycloak token is required"},
                status=status.HTTP_400_BAD_REQUEST,
            )

        # Validate the token and get user info
        logger.debug(f"Validating token of length: {len(keycloak_token)}")
        user_info = keycloak_manager.validate_token(keycloak_token)
        if not user_info:
            logger.warning("Token validation failed")
            return Response(
                {"error": "Invalid or expired token"},
                status=status.HTTP_401_UNAUTHORIZED,
            )

        # Ensure we have a subject ID in the token
        if not user_info.get("sub"):
            logger.warning("Token validation succeeded but missing subject ID")
            return Response(
                {"error": "Token missing required user information"},
                status=status.HTTP_401_UNAUTHORIZED,
            )

        # Get user roles from the token
        roles = keycloak_manager.get_user_roles(keycloak_token)

        # Get user organizations from the token
        organizations = keycloak_manager.get_user_organizations(keycloak_token)

        # Sync the user information with our database
        logger.info(f"Syncing user with Keycloak ID: {user_info.get('sub')}")
        user = keycloak_manager.sync_user_from_keycloak(user_info, roles, organizations)
        if not user:
            logger.error(
                f"Failed to sync user with Keycloak ID: {user_info.get('sub')}"
            )
            return Response(
                {"error": "Failed to synchronize user information"},
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

        # Create Django tokens for the user
        refresh = RefreshToken.for_user(user)

        # Get user's organizations and their roles
        organizations = AuthorizationService.get_user_organizations(user.id)

        # Get user's dataset-specific permissions
        datasets = AuthorizationService.get_user_datasets(user.id)

        return Response(
            {
                "refresh": str(refresh),
                "access": str(refresh.access_token),
                "user": {
                    "id": user.id,
                    "username": user.username,
                    "email": user.email,
                    "first_name": user.first_name,
                    "last_name": user.last_name,
                    "is_staff": user.is_staff,
                    "is_superuser": user.is_superuser,
                    "organizations": organizations,
                    "datasets": datasets,
                },
            }
        )


class UserInfoView(views.APIView):
    """
    View for getting the current user's information.
    """

    permission_classes = [IsAuthenticated]

    def get(self, request: Request) -> Response:
        import structlog

        logger = structlog.getLogger(__name__)

        user = request.user
        logger.debug(
            f"Getting user info for user: {user.username if hasattr(user, 'username') else 'anonymous'}"
        )

        try:
            # Get user's organizations and their roles
            organizations = AuthorizationService.get_user_organizations(user.id)  # type: ignore[arg-type]

            # Get user's dataset-specific permissions
            datasets = AuthorizationService.get_user_datasets(user.id)  # type: ignore[arg-type]
        except Exception as e:
            logger.error(f"Error retrieving user information: {str(e)}")
            return Response(
                {"error": "Failed to retrieve user information"},
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

        return Response(
            {
                "id": user.id,
                "username": user.username,
                "email": user.email,  # type: ignore[union-attr]
                "first_name": user.first_name,  # type: ignore[union-attr]
                "last_name": user.last_name,  # type: ignore[union-attr]
                "is_staff": user.is_staff,
                "is_superuser": user.is_superuser,
                "organizations": organizations,
                "datasets": datasets,
            }
        )


class UserConsentView(APIView):
    """
    API view for managing user consent for activity tracking.
    """

    permission_classes = [IsAuthenticated]

    def get(self, request: Request) -> Response:
        """
        Get the current user's consent settings.
        """
        try:
            consent = UserConsent.objects.get(user=request.user)  # type: ignore[misc]
        except UserConsent.DoesNotExist:
            # Create a default consent object with tracking disabled
            consent = UserConsent.objects.create(
                user=request.user, activity_tracking_enabled=False  # type: ignore[misc]
            )

        serializer = UserConsentSerializer(consent)
        return Response(serializer.data)

    def put(self, request: Request) -> Response:
        """
        Update the current user's consent settings.
        """
        try:
            consent = UserConsent.objects.get(user=request.user)  # type: ignore[misc]
        except UserConsent.DoesNotExist:
            consent = UserConsent.objects.create(
                user=request.user, activity_tracking_enabled=False  # type: ignore[misc]
            )

        serializer = UserConsentSerializer(
            consent, data=request.data, context={"request": request}
        )
        if serializer.is_valid():
            serializer.save()
            return Response(serializer.data)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)



"""GraphQL schema for AI Model."""

# mypy: disable-error-code=union-attr

import datetime
from typing import List, Optional

import strawberry
import strawberry_django
from django.core.exceptions import ValidationError as DjangoValidationError
from strawberry.types import Info
from strawberry_django.pagination import OffsetPaginationInput

from api.models.AIModel import AIModel, ModelAPIKey, ModelEndpoint
from api.models.Dataset import Tag
from api.schema.base_mutation import BaseMutation, MutationResponse
from api.schema.extensions import TrackActivity, TrackModelActivity
from api.types.type_aimodel import (
    AIModelFilter,
    AIModelOrder,
    AIModelProviderEnum,
    AIModelStatusEnum,
    AIModelTypeEnum,
    EndpointAuthTypeEnum,
    EndpointHTTPMethodEnum,
    TypeAIModel,
    TypeModelEndpoint,
)
from api.utils.graphql_telemetry import trace_resolver
from authorization.graphql_permissions import IsAuthenticated
from authorization.models import OrganizationMembership, Role


@trace_resolver(name="update_aimodel_tags", attributes={"component": "aimodel"})
def _update_aimodel_tags(model: AIModel, tags: Optional[List[str]]) -> None:
    """Update tags for an AI model."""
    if tags is None:
        return
    model.tags.clear()
    for tag in tags:
        model.tags.add(
            Tag.objects.get_or_create(defaults={"value": tag}, value__iexact=tag)[0]
        )
    model.save()


def _update_aimodel_sectors(model: AIModel, sectors: List[str]) -> None:
    """Helper function to update sectors for an AI model."""
    from api.models import Sector

    model.sectors.clear()
    for sector_name in sectors:
        try:
            sector = Sector.objects.get(name__iexact=sector_name)
            model.sectors.add(sector)
        except Sector.DoesNotExist:
            pass
    model.save()


def _update_aimodel_geographies(model: AIModel, geographies: List[str]) -> None:
    """Helper function to update geographies for an AI model."""
    from api.models import Geography

    model.geographies.clear()
    for geography_name in geographies:
        try:
            geography = Geography.objects.get(name__iexact=geography_name)
            model.geographies.add(geography)
        except Geography.DoesNotExist:
            pass
    model.save()


@strawberry.input
class CreateAIModelInput:
    """Input for creating a new AI Model."""

    name: str
    display_name: str
    description: str
    model_type: AIModelTypeEnum
    provider: AIModelProviderEnum
    version: Optional[str] = None
    provider_model_id: Optional[str] = None
    supports_streaming: bool = False
    max_tokens: Optional[int] = None
    supported_languages: Optional[List[str]] = None
    input_schema: Optional[strawberry.scalars.JSON] = None
    output_schema: Optional[strawberry.scalars.JSON] = None
    tags: Optional[List[str]] = None
    sectors: Optional[List[str]] = None
    geographies: Optional[List[str]] = None
    metadata: Optional[strawberry.scalars.JSON] = None
    is_public: bool = False


@strawberry.input
class UpdateAIModelInput:
    """Input for updating an AI Model."""

    id: int
    name: Optional[str] = None
    display_name: Optional[str] = None
    description: Optional[str] = None
    model_type: Optional[AIModelTypeEnum] = None
    provider: Optional[AIModelProviderEnum] = None
    version: Optional[str] = None
    provider_model_id: Optional[str] = None
    supports_streaming: Optional[bool] = None
    max_tokens: Optional[int] = None
    supported_languages: Optional[List[str]] = None
    input_schema: Optional[strawberry.scalars.JSON] = None
    output_schema: Optional[strawberry.scalars.JSON] = None
    tags: Optional[List[str]] = None
    sectors: Optional[List[str]] = None
    geographies: Optional[List[str]] = None
    metadata: Optional[strawberry.scalars.JSON] = None
    is_public: Optional[bool] = None
    is_active: Optional[bool] = None
    status: Optional[AIModelStatusEnum] = None


@strawberry.input
class CreateModelEndpointInput:
    """Input for creating a model endpoint."""

    model_id: int
    url: str
    http_method: EndpointHTTPMethodEnum = EndpointHTTPMethodEnum.POST
    auth_type: EndpointAuthTypeEnum = EndpointAuthTypeEnum.BEARER
    auth_header_name: str = "Authorization"
    headers: Optional[strawberry.scalars.JSON] = None
    request_template: Optional[strawberry.scalars.JSON] = None
    response_path: Optional[str] = None
    timeout_seconds: int = 30
    max_retries: int = 3
    is_primary: bool = True
    is_active: bool = True
    rate_limit_per_minute: Optional[int] = None


@strawberry.input
class UpdateModelEndpointInput:
    """Input for updating a model endpoint."""

    id: int
    url: Optional[str] = None
    http_method: Optional[EndpointHTTPMethodEnum] = None
    auth_type: Optional[EndpointAuthTypeEnum] = None
    auth_header_name: Optional[str] = None
    headers: Optional[strawberry.scalars.JSON] = None
    request_template: Optional[strawberry.scalars.JSON] = None
    response_path: Optional[str] = None
    timeout_seconds: Optional[int] = None
    max_retries: Optional[int] = None
    is_primary: Optional[bool] = None
    is_active: Optional[bool] = None
    rate_limit_per_minute: Optional[int] = None


@strawberry.type
class Query:
    """Queries for AI Models."""

    @strawberry_django.field(
        filters=AIModelFilter,
        pagination=True,
        order=AIModelOrder,
    )
    @trace_resolver(name="ai_models", attributes={"component": "aimodel"})
    def ai_models(
        self,
        info: Info,
        filters: Optional[AIModelFilter] = strawberry.UNSET,
        pagination: Optional[OffsetPaginationInput] = strawberry.UNSET,
        order: Optional[AIModelOrder] = strawberry.UNSET,
    ) -> List[TypeAIModel]:
        """Get all AI models."""
        organization = info.context.context.get("organization")
        user = info.context.user

        if organization:
            queryset = AIModel.objects.filter(organization=organization)
        else:
            # If user is authenticated
            if user.is_authenticated:
                # If user is superuser, show all models
                if user.is_superuser:
                    queryset = AIModel.objects.all()
                else:
                    # For authenticated users, show their models and public models
                    queryset = AIModel.objects.filter(
                        user=user
                    ) | AIModel.objects.filter(is_public=True, is_active=True)
            else:
                # For non-authenticated users, only show public active models
                queryset = AIModel.objects.filter(is_public=True, is_active=True)

        if filters is not strawberry.UNSET:
            queryset = strawberry_django.filters.apply(filters, queryset, info)

        if order is not strawberry.UNSET:
            queryset = strawberry_django.ordering.apply(order, queryset, info)

        if pagination is not strawberry.UNSET:
            queryset = strawberry_django.pagination.apply(pagination, queryset)

        return TypeAIModel.from_django_list(list(queryset.distinct()))

    @strawberry.field
    @trace_resolver(name="get_ai_model", attributes={"component": "aimodel"})
    def get_ai_model(self, info: Info, model_id: int) -> Optional[TypeAIModel]:
        """Get an AI model by ID."""
        user = info.context.user
        try:
            model = AIModel.objects.get(id=model_id)

            # Check permissions
            if model.is_public and model.is_active:
                return TypeAIModel.from_django(model)

            if not user.is_authenticated:
                return None

            if user.is_superuser or model.user == user:
                return TypeAIModel.from_django(model)

            if model.organization:
                org_member = OrganizationMembership.objects.filter(
                    user=user, organization=model.organization
                ).first()
                if org_member and org_member.role.can_view:
                    return TypeAIModel.from_django(model)

            return None
        except AIModel.DoesNotExist:
            return None

    @strawberry.field
    @trace_resolver(name="get_model_endpoints", attributes={"component": "aimodel"})
    def get_model_endpoints(self, info: Info, model_id: int) -> List[TypeModelEndpoint]:
        """Get all endpoints for an AI model."""
        user = info.context.user
        try:
            model = AIModel.objects.get(id=model_id)

            # Check permissions
            if not model.is_public and not user.is_authenticated:
                return []

            if not model.is_public and model.user != user and not user.is_superuser:
                if model.organization:
                    org_member = OrganizationMembership.objects.filter(
                        user=user, organization=model.organization
                    ).first()
                    if not org_member or not org_member.role.can_view:
                        return []
                else:
                    return []

            endpoints = ModelEndpoint.objects.filter(model=model)
            return TypeModelEndpoint.from_django_list(list(endpoints))
        except AIModel.DoesNotExist:
            return []


@strawberry.type
class Mutation:
    """Mutations for AI Models."""

    @strawberry.mutation
    @BaseMutation.mutation(
        permission_classes=[IsAuthenticated],
        trace_name="create_ai_model",
        trace_attributes={"component": "aimodel"},
        track_activity={
            "verb": "created",
            "get_data": lambda result, **kwargs: {
                "model_id": str(result.id),
                "model_name": result.name,
                "organization": (
                    str(result.organization.id) if result.organization else None
                ),
            },
        },
    )
    def create_ai_model(
        self, info: Info, input: CreateAIModelInput
    ) -> MutationResponse[TypeAIModel]:
        """Create a new AI model."""
        organization = info.context.context.get("organization")
        user = info.context.user

        # Prepare supported_languages
        supported_languages = input.supported_languages or []

        # Prepare schemas
        input_schema = input.input_schema or {}
        output_schema = input.output_schema or {}

        # Prepare metadata
        metadata = input.metadata or {}

        try:
            model = AIModel.objects.create(
                name=input.name,
                display_name=input.display_name,
                version=input.version or "",
                description=input.description,
                model_type=input.model_type,
                provider=input.provider,
                provider_model_id=input.provider_model_id or "",
                organization=organization,
                user=user,
                supports_streaming=input.supports_streaming,
                max_tokens=input.max_tokens,
                supported_languages=supported_languages,
                input_schema=input_schema,
                output_schema=output_schema,
                metadata=metadata,
                is_public=input.is_public,
                status="REGISTERED",
            )
            # Handle tags separately after model creation
            if input.tags is not None:
                _update_aimodel_tags(model, input.tags)

            # Handle sectors
            if input.sectors is not None:
                _update_aimodel_sectors(model, input.sectors)

            # Handle geographies
            if input.geographies is not None:
                _update_aimodel_geographies(model, input.geographies)

            return MutationResponse.success_response(TypeAIModel.from_django(model))
        except Exception as e:
            raise DjangoValidationError(f"Failed to create AI model: {str(e)}")

    @strawberry.mutation
    @BaseMutation.mutation(
        permission_classes=[IsAuthenticated],
        trace_name="update_ai_model",
        trace_attributes={"component": "aimodel"},
        track_activity={
            "verb": "updated",
            "get_data": lambda result, **kwargs: {
                "model_id": str(result.id),
                "model_name": result.name,
                "organization": (
                    str(result.organization.id) if result.organization else None
                ),
            },
        },
    )
    def update_ai_model(
        self, info: Info, input: UpdateAIModelInput
    ) -> MutationResponse[TypeAIModel]:
        """Update an AI model."""
        user = info.context.user

        try:
            model = AIModel.objects.get(id=input.id)
        except AIModel.DoesNotExist:
            raise DjangoValidationError(f"AI Model with ID {input.id} does not exist.")

        # Check permissions
        if not user.is_superuser and model.user != user:
            if model.organization:
                org_member = OrganizationMembership.objects.filter(
                    user=user, organization=model.organization
                ).first()
                if not org_member or not org_member.role.can_change:
                    raise DjangoValidationError(
                        "You don't have permission to update this model."
                    )
            else:
                raise DjangoValidationError(
                    "You don't have permission to update this model."
                )

        # Update fields
        if input.name is not None:
            model.name = input.name
        if input.display_name is not None:
            model.display_name = input.display_name
        if input.description is not None:
            model.description = input.description
        if input.version is not None:
            model.version = input.version
        if input.provider_model_id is not None:
            model.provider_model_id = input.provider_model_id
        if input.supports_streaming is not None:
            model.supports_streaming = input.supports_streaming
        if input.max_tokens is not None:
            model.max_tokens = input.max_tokens
        if input.supported_languages is not None:
            model.supported_languages = input.supported_languages
        if input.input_schema is not None:
            model.input_schema = input.input_schema
        if input.output_schema is not None:
            model.output_schema = input.output_schema
        if input.metadata is not None:
            model.metadata = input.metadata
        if input.is_public is not None:
            model.is_public = input.is_public
        if input.is_active is not None:
            model.is_active = input.is_active
        if input.status is not None:
            model.status = input.status
        if input.model_type is not None:
            model.model_type = input.model_type
        if input.provider is not None:
            model.provider = input.provider

        model.save()

        # Handle tags separately
        if input.tags is not None:
            _update_aimodel_tags(model, input.tags)

        # Handle sectors
        if input.sectors is not None:
            _update_aimodel_sectors(model, input.sectors)

        # Handle geographies
        if input.geographies is not None:
            _update_aimodel_geographies(model, input.geographies)

        return MutationResponse.success_response(TypeAIModel.from_django(model))

    @strawberry.mutation
    @BaseMutation.mutation(
        permission_classes=[IsAuthenticated],
        trace_name="delete_ai_model",
        trace_attributes={"component": "aimodel"},
        track_activity={
            "verb": "deleted",
            "get_data": lambda result, **kwargs: {
                "model_id": str(kwargs.get("model_id")),
                "success": result,
            },
        },
    )
    def delete_ai_model(self, info: Info, model_id: int) -> MutationResponse[bool]:
        """Delete an AI model."""
        user = info.context.user

        try:
            model = AIModel.objects.get(id=model_id)
        except AIModel.DoesNotExist:
            raise DjangoValidationError(f"AI Model with ID {model_id} does not exist.")

        # Check permissions
        if not user.is_superuser and model.user != user:
            if model.organization:
                org_member = OrganizationMembership.objects.filter(
                    user=user, organization=model.organization
                ).first()
                if not org_member or not org_member.role.can_delete:
                    raise DjangoValidationError(
                        "You don't have permission to delete this model."
                    )
            else:
                raise DjangoValidationError(
                    "You don't have permission to delete this model."
                )

        model.delete()
        return MutationResponse.success_response(True)

    @strawberry.mutation
    @BaseMutation.mutation(
        permission_classes=[IsAuthenticated],
        trace_name="create_model_endpoint",
        trace_attributes={"component": "aimodel"},
        track_activity={
            "verb": "created endpoint",
            "get_data": lambda result, **kwargs: {
                "endpoint_id": str(result.id),
                "model_id": str(result.model.id),
            },
        },
    )
    def create_model_endpoint(
        self, info: Info, input: CreateModelEndpointInput
    ) -> MutationResponse[TypeModelEndpoint]:
        """Create a new model endpoint."""
        user = info.context.user

        try:
            model = AIModel.objects.get(id=input.model_id)
        except AIModel.DoesNotExist:
            raise DjangoValidationError(
                f"AI Model with ID {input.model_id} does not exist."
            )

        # Check permissions
        if not user.is_superuser and model.user != user:
            if model.organization:
                org_member = OrganizationMembership.objects.filter(
                    user=user, organization=model.organization
                ).first()
                if not org_member or not org_member.role.can_change:
                    raise DjangoValidationError(
                        "You don't have permission to add endpoints to this model."
                    )
            else:
                raise DjangoValidationError(
                    "You don't have permission to add endpoints to this model."
                )

        # If this is primary, unset other primary endpoints
        if input.is_primary:
            ModelEndpoint.objects.filter(model=model, is_primary=True).update(
                is_primary=False
            )

        endpoint = ModelEndpoint.objects.create(
            model=model,
            url=input.url,
            http_method=input.http_method,
            auth_type=input.auth_type,
            auth_header_name=input.auth_header_name,
            headers=input.headers or {},
            request_template=input.request_template or {},
            response_path=input.response_path or "",
            timeout_seconds=input.timeout_seconds,
            max_retries=input.max_retries,
            is_primary=input.is_primary,
            rate_limit_per_minute=input.rate_limit_per_minute,
            is_active=input.is_active,
        )

        return MutationResponse.success_response(
            TypeModelEndpoint.from_django(endpoint)
        )

    @strawberry.mutation
    @BaseMutation.mutation(
        permission_classes=[IsAuthenticated],
        trace_name="update_model_endpoint",
        trace_attributes={"component": "aimodel"},
        track_activity={
            "verb": "updated endpoint",
            "get_data": lambda result, **kwargs: {
                "endpoint_id": str(result.id),
                "model_id": str(result.model.id),
            },
        },
    )
    def update_model_endpoint(
        self, info: Info, input: UpdateModelEndpointInput
    ) -> MutationResponse[TypeModelEndpoint]:
        """Update a model endpoint."""
        user = info.context.user

        try:
            endpoint = ModelEndpoint.objects.get(id=input.id)
        except ModelEndpoint.DoesNotExist:
            raise DjangoValidationError(
                f"Model Endpoint with ID {input.id} does not exist."
            )

        model = endpoint.model

        # Check permissions
        if not user.is_superuser and model.user != user:
            if model.organization:
                org_member = OrganizationMembership.objects.filter(
                    user=user, organization=model.organization
                ).first()
                if not org_member or not org_member.role.can_change:
                    raise DjangoValidationError(
                        "You don't have permission to update this endpoint."
                    )
            else:
                raise DjangoValidationError(
                    "You don't have permission to update this endpoint."
                )

        # Update fields
        if input.url is not None:
            endpoint.url = input.url
        if input.http_method is not None:
            endpoint.http_method = input.http_method
        if input.auth_type is not None:
            endpoint.auth_type = input.auth_type
        if input.auth_header_name is not None:
            endpoint.auth_header_name = input.auth_header_name
        if input.headers is not None:
            endpoint.headers = input.headers
        if input.request_template is not None:
            endpoint.request_template = input.request_template
        if input.response_path is not None:
            endpoint.response_path = input.response_path
        if input.timeout_seconds is not None:
            endpoint.timeout_seconds = input.timeout_seconds
        if input.max_retries is not None:
            endpoint.max_retries = input.max_retries
        if input.is_active is not None:
            endpoint.is_active = input.is_active
        if input.rate_limit_per_minute is not None:
            endpoint.rate_limit_per_minute = input.rate_limit_per_minute

        # If setting as primary, unset other primary endpoints
        if input.is_primary is not None and input.is_primary:
            ModelEndpoint.objects.filter(model=model, is_primary=True).exclude(
                id=endpoint.id
            ).update(is_primary=False)
            endpoint.is_primary = True

        endpoint.save()
        return MutationResponse.success_response(
            TypeModelEndpoint.from_django(endpoint)
        )

    @strawberry.mutation
    @BaseMutation.mutation(
        permission_classes=[IsAuthenticated],
        trace_name="delete_model_endpoint",
        trace_attributes={"component": "aimodel"},
        track_activity={
            "verb": "deleted endpoint",
            "get_data": lambda result, **kwargs: {
                "endpoint_id": str(kwargs.get("endpoint_id")),
                "success": result,
            },
        },
    )
    def delete_model_endpoint(
        self, info: Info, endpoint_id: int
    ) -> MutationResponse[bool]:
        """Delete a model endpoint."""
        user = info.context.user

        try:
            endpoint = ModelEndpoint.objects.get(id=endpoint_id)
        except ModelEndpoint.DoesNotExist:
            raise DjangoValidationError(
                f"Model Endpoint with ID {endpoint_id} does not exist."
            )

        model = endpoint.model

        # Check permissions
        if not user.is_superuser and model.user != user:
            if model.organization:
                org_member = OrganizationMembership.objects.filter(
                    user=user, organization=model.organization
                ).first()
                if not org_member or not org_member.role.can_delete:
                    raise DjangoValidationError(
                        "You don't have permission to delete this endpoint."
                    )
            else:
                raise DjangoValidationError(
                    "You don't have permission to delete this endpoint."
                )

        endpoint.delete()
        return MutationResponse.success_response(True)



"""Search view for AIModel using Elasticsearch."""

from typing import Any, Dict, List, Optional, Tuple, Union, cast

import structlog
from elasticsearch_dsl import Q as ESQ
from elasticsearch_dsl import Search
from elasticsearch_dsl.query import Query as ESQuery
from rest_framework import serializers
from rest_framework.permissions import AllowAny

from api.models.AIModel import AIModel
from api.utils.telemetry_utils import trace_method
from api.views.paginated_elastic_view import PaginatedElasticSearchAPIView
from search.documents import AIModelDocument

logger = structlog.get_logger(__name__)


class AIModelDocumentSerializer(serializers.ModelSerializer):
    """Serializer for AIModel document."""

    tags = serializers.ListField()
    supported_languages = serializers.ListField()
    is_individual_model = serializers.BooleanField()
    has_active_endpoints = serializers.BooleanField()
    endpoint_count = serializers.IntegerField()

    class OrganizationSerializer(serializers.Serializer):
        name = serializers.CharField()
        logo = serializers.CharField()

    class UserSerializer(serializers.Serializer):
        name = serializers.CharField()
        bio = serializers.CharField()
        profile_picture = serializers.CharField()

    class EndpointSerializer(serializers.Serializer):
        url = serializers.CharField()
        http_method = serializers.CharField()
        auth_type = serializers.CharField()
        is_primary = serializers.BooleanField()
        is_active = serializers.BooleanField()

    organization = OrganizationSerializer(allow_null=True)
    user = UserSerializer(allow_null=True)
    endpoints = EndpointSerializer(many=True)

    class Meta:
        model = AIModel
        fields = [
            "id",
            "name",
            "display_name",
            "description",
            "version",
            "model_type",
            "provider",
            "provider_model_id",
            "status",
            "is_public",
            "is_active",
            "tags",
            "supported_languages",
            "supports_streaming",
            "max_tokens",
            "average_latency_ms",
            "success_rate",
            "last_audit_score",
            "audit_count",
            "created_at",
            "updated_at",
            "last_tested_at",
            "is_individual_model",
            "has_active_endpoints",
            "endpoint_count",
            "organization",
            "user",
            "endpoints",
        ]


class SearchAIModel(PaginatedElasticSearchAPIView):
    """View for searching AI models."""

    serializer_class = AIModelDocumentSerializer
    document_class = AIModelDocument
    permission_classes = [AllowAny]

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.searchable_fields: List[str]
        self.aggregations: Dict[str, str]
        self.searchable_fields, self.aggregations = (
            self.get_searchable_and_aggregations()
        )
        self.logger = structlog.get_logger(__name__)

    @trace_method(
        name="get_searchable_and_aggregations",
        attributes={"component": "search_aimodel"},
    )
    def get_searchable_and_aggregations(self) -> Tuple[List[str], Dict[str, str]]:
        """Get searchable fields and aggregations for the search."""
        searchable_fields: List[str] = [
            "name",
            "display_name",
            "description",
            "tags",
            "provider_model_id",
        ]

        aggregations: Dict[str, str] = {
            "model_type": "terms",
            "provider": "terms",
            "status": "terms",
            "tags.raw": "terms",
            "supported_languages": "terms",
            "is_public": "terms",
            "is_active": "terms",
            "supports_streaming": "terms",
        }

        return searchable_fields, aggregations

    @trace_method(name="add_aggregations", attributes={"component": "search_aimodel"})
    def add_aggregations(self, search: Search) -> Search:
        """Add aggregations to the search query."""
        for aggregation_field in self.aggregations:
            search.aggs.bucket(
                aggregation_field.replace(".raw", ""),
                self.aggregations[aggregation_field],
                field=aggregation_field,
            )
        return search

    @trace_method(
        name="generate_q_expression", attributes={"component": "search_aimodel"}
    )
    def generate_q_expression(
        self, query: str
    ) -> Optional[Union[ESQuery, List[ESQuery]]]:
        """Generate Elasticsearch Query expression."""
        if query:
            queries: List[ESQuery] = []
            for field in self.searchable_fields:
                queries.append(
                    ESQ("fuzzy", **{field: {"value": query, "fuzziness": "AUTO"}})
                )
        else:
            queries = [ESQ("match_all")]

        return ESQ("bool", should=queries, minimum_should_match=1)

    @trace_method(name="add_filters", attributes={"component": "search_aimodel"})
    def add_filters(self, filters: Dict[str, str], search: Search) -> Search:
        """Add filters to the search query."""
        for filter_key in filters:
            if filter_key in ["tags"]:
                # Handle multi-value filters
                raw_filter = filter_key + ".raw"
                if raw_filter in self.aggregations:
                    filter_values = filters[filter_key].split(",")
                    search = search.filter("terms", **{raw_filter: filter_values})
                else:
                    search = search.filter("term", **{filter_key: filters[filter_key]})
            elif filter_key in [
                "model_type",
                "provider",
                "status",
                "supported_languages",
            ]:
                # Handle single or multi-value filters
                filter_values = filters[filter_key].split(",")
                if len(filter_values) > 1:
                    search = search.filter("terms", **{filter_key: filter_values})
                else:
                    search = search.filter("term", **{filter_key: filters[filter_key]})
            elif filter_key in ["is_public", "is_active", "supports_streaming"]:
                # Handle boolean filters
                bool_value = filters[filter_key].lower() in ["true", "1", "yes"]
                search = search.filter("term", **{filter_key: bool_value})
            elif filter_key == "organization":
                # Filter by organization name
                search = search.filter(
                    "nested",
                    path="organization",
                    query={"term": {"organization.name": filters[filter_key]}},
                )
            elif filter_key == "user":
                # Filter by user name
                search = search.filter(
                    "nested",
                    path="user",
                    query={"term": {"user.name": filters[filter_key]}},
                )

        return search

    @trace_method(name="add_sort", attributes={"component": "search_aimodel"})
    def add_sort(self, sort: str, search: Search, order: str) -> Search:
        """Add sorting to the search query."""
        if sort == "alphabetical":
            search = search.sort({"name.raw": {"order": order}})
        elif sort == "recent":
            search = search.sort({"updated_at": {"order": order}})
        elif sort == "created":
            search = search.sort({"created_at": {"order": order}})
        elif sort == "performance":
            # Sort by success_rate, then by average_latency_ms (lower is better)
            search = search.sort(
                {"success_rate": {"order": "desc"}},
                {"average_latency_ms": {"order": "asc"}},
            )
        elif sort == "audit_score":
            search = search.sort({"last_audit_score": {"order": order}})
        elif sort == "popular":
            # Sort by audit_count as a proxy for usage
            search = search.sort({"audit_count": {"order": order}})

        return search



from typing import TYPE_CHECKING, Annotated, List, Optional

import strawberry
import strawberry_django
from strawberry import auto

from api.types.base_type import BaseType
from authorization.models import OrganizationMembership, Role, User

if TYPE_CHECKING:
    from authorization.types import TypeOrganizationMembership


@strawberry_django.filter(User)
class UserFilter:
    """Filter for user."""

    id: Optional[strawberry.ID]
    username: Optional[str]
    email: Optional[str]


@strawberry_django.order(User)
class UserOrder:
    """Order for user."""

    username: auto
    first_name: auto
    last_name: auto
    date_joined: auto


@strawberry_django.type(
    User,
    fields=[
        "id",
        "username",
        "email",
        "first_name",
        "last_name",
        "bio",
        "profile_picture",
        "date_joined",
        "last_login",
    ],
    filters=UserFilter,
    pagination=True,
    order=True,  # Changed from UserOrder to True
)
class TypeUser(BaseType):
    """Type for user."""

    @strawberry.field
    def organization_memberships(
        self,
    ) -> List[
        Annotated["TypeOrganizationMembership", strawberry.lazy("authorization.types")]
    ]:
        """Get organization memberships for this user."""
        try:
            from authorization.types import TypeOrganizationMembership

            user_id = str(getattr(self, "id", ""))
            queryset = OrganizationMembership.objects.filter(user_id=user_id)
            return TypeOrganizationMembership.from_django_list(queryset)
        except Exception:
            return []

    @strawberry.field
    def full_name(self) -> str:
        """Get full name of the user."""
        first_name = getattr(self, "first_name", "")
        last_name = getattr(self, "last_name", "")
        if first_name and last_name:
            return f"{first_name} {last_name}"
        elif first_name:
            return first_name
        elif last_name:
            return last_name
        else:
            return getattr(self, "username", "")



import abc
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union

from django.core.cache import cache
from django.http import HttpRequest, HttpResponse
from elasticsearch_dsl import Search
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_framework.views import APIView

from api.signals.dataset_signals import SEARCH_CACHE_VERSION_KEY

T = TypeVar("T")
SearchType = TypeVar("SearchType", bound=Search)
SerializerType = TypeVar("SerializerType", bound=Serializer)


class PaginatedElasticSearchAPIView(Generic[SerializerType, SearchType], APIView):
    """Base class for paginated Elasticsearch API views."""

    serializer_class: Type[SerializerType]
    document_class: Type[SearchType]
    permission_classes = [AllowAny]  # Allow unauthenticated access by default

    @abc.abstractmethod
    def generate_q_expression(self, query: str) -> Any:
        """This method should be overridden and return a Q() expression."""
        pass

    @abc.abstractmethod
    def add_aggregations(self, search: SearchType) -> SearchType:
        """This method should be overridden and return a Search object with aggregations added."""
        pass

    @abc.abstractmethod
    def add_filters(self, filters: Dict[str, Any], search: SearchType) -> SearchType:
        """This method should be overridden and return a Search object with filters added."""
        pass

    @abc.abstractmethod
    def add_sort(self, sort: str, search: SearchType, order: str) -> SearchType:
        """This method should be overridden and return a Search object with sort added."""
        pass

    def get_search(self) -> SearchType:
        """Get search instance."""
        if hasattr(self.document_class, "search"):
            return self.document_class.search()  # type: ignore
        raise AttributeError(
            f"{self.document_class.__name__} does not have a search method"
        )

    def get(self, request: HttpRequest) -> Response:
        """Handle GET request and return paginated search results."""
        try:
            # Generate cache key based on request parameters
            cache_key = self._generate_cache_key(request)
            cached_result: Optional[Dict[str, Any]] = cache.get(cache_key)

            # TODO: Fix cache issues on different model updates
            # if cached_result:
            #     return Response(cached_result)

            # Original search logic
            query: str = request.GET.get("query", "")
            page: int = int(request.GET.get("page", 1))
            size: int = int(request.GET.get("size", 10))
            sort: str = request.GET.get("sort", "alphabetical")
            order: str = request.GET.get("order", "asc")
            # Handle multiple values for the same filter parameter
            filters: Dict[str, Any] = {}
            for key, values in request.GET.lists():
                if key not in ["query", "page", "size", "sort", "order"]:
                    if len(values) > 1:
                        # Multiple values: join with comma for OR filtering
                        filters[key] = ",".join(values)
                    else:
                        # Single value - but check if it's already comma-separated
                        single_value = values[0]
                        filters[key] = single_value

            q = self.generate_q_expression(query)
            search = self.get_search().query(q)
            search = self.add_aggregations(search)
            search = self.add_filters(filters, search)
            search = self.add_sort(sort, search, order)
            search = search[(page - 1) * size : page * size]
            response = search.execute()

            serializer = self.serializer_class(response, many=True)
            aggregations: Dict[str, Any] = response.aggregations.to_dict()

            if "metadata" in aggregations:

                metadata_aggregations = aggregations["metadata"]["filtered_metadata"][
                    "composite_agg"
                ]["buckets"]
                aggregations.pop("metadata")
                for agg in metadata_aggregations:
                    label: str = agg["key"]["metadata_label"]
                value: str = agg["key"].get("metadata_value", "")
                if label not in aggregations:
                    aggregations[label] = {}
                aggregations[label][value] = agg["doc_count"]

            if "catalogs" in aggregations:
                aggregations.pop("catalogs")
            # Handle sectors aggregation (now comes as "sectors.raw")
            if "sectors.raw" in aggregations:
                sectors_agg = aggregations["sectors.raw"]["buckets"]
                aggregations.pop("sectors.raw")
                aggregations["sectors"] = {}
                for agg in sectors_agg:
                    aggregations["sectors"][agg["key"]] = agg["doc_count"]
            elif "sectors" in aggregations:
                sectors_agg = aggregations["sectors"]["buckets"]
                aggregations.pop("sectors")
                aggregations["sectors"] = {}
                for agg in sectors_agg:
                    aggregations["sectors"][agg["key"]] = agg["doc_count"]

            # Handle tags aggregation (now comes as "tags.raw")
            if "tags.raw" in aggregations:
                tags_agg = aggregations["tags.raw"]["buckets"]
                aggregations.pop("tags.raw")
                aggregations["tags"] = {}
                for agg in tags_agg:
                    aggregations["tags"][agg["key"]] = agg["doc_count"]
            elif "tags" in aggregations:
                tags_agg = aggregations["tags"]["buckets"]
                aggregations.pop("tags")
                aggregations["tags"] = {}
                for agg in tags_agg:
                    aggregations["tags"][agg["key"]] = agg["doc_count"]

            if "formats" in aggregations:
                formats_agg = aggregations["formats"]["buckets"]
                aggregations.pop("formats")
                aggregations["formats"] = {}
                for agg in formats_agg:
                    aggregations["formats"][agg["key"]] = agg["doc_count"]

            # Handle geographies aggregation (now comes as "geographies.raw")
            if "geographies.raw" in aggregations:
                geographies_agg = aggregations["geographies.raw"]["buckets"]
                aggregations.pop("geographies.raw")
                aggregations["geographies"] = {}
                for agg in geographies_agg:
                    aggregations["geographies"][agg["key"]] = agg["doc_count"]
            elif "geographies" in aggregations:
                geographies_agg = aggregations["geographies"]["buckets"]
                aggregations.pop("geographies")
                aggregations["geographies"] = {}
                for agg in geographies_agg:
                    aggregations["geographies"][agg["key"]] = agg["doc_count"]

            if "status" in aggregations:
                status_agg = aggregations["status"]["buckets"]
                aggregations.pop("status")
                aggregations["status"] = {}
                for agg in status_agg:
                    aggregations["status"][agg["key"]] = agg["doc_count"]

            if "running_status" in aggregations:
                running_status_agg = aggregations["running_status"]["buckets"]
                aggregations.pop("running_status")
                aggregations["running_status"] = {}
                for agg in running_status_agg:
                    aggregations["running_status"][agg["key"]] = agg["doc_count"]

            if "is_individual_usecase" in aggregations:
                is_individual_usecase_agg = aggregations["is_individual_usecase"][
                    "buckets"
                ]
                aggregations.pop("is_individual_usecase")
                aggregations["is_individual_usecase"] = {}
                for agg in is_individual_usecase_agg:
                    aggregations["is_individual_usecase"][agg["key"]] = agg["doc_count"]

            result: Dict[str, Any] = {
                "results": serializer.data,
                "total": response.hits.total.value,  # type: ignore
                "aggregations": aggregations,
            }

            # Cache the result
            cache.set(cache_key, result, timeout=3600)  # Cache for 1 hour

            return Response(result)
        except Exception as e:
            return Response(str(e), status=500)

    def _generate_cache_key(self, request: HttpRequest) -> str:
        """Generate a unique cache key based on request parameters and cache version."""
        params: Dict[str, str] = {
            "query": request.GET.get("query", ""),
            "page": request.GET.get("page", "1"),
            "size": request.GET.get("size", "10"),
            "sort": request.GET.get("sort", "alphabetical"),
            "filters": str(sorted(request.GET.dict().items())),
            "version": str(cache.get(SEARCH_CACHE_VERSION_KEY, 0)),
        }
        return f"search_results:{hash(frozenset(params.items()))}"



import functools
from typing import Any, Callable, Optional, TypeVar, cast

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import Model
from django.http import HttpRequest

from authorization.activity import record_activity
from authorization.consent import UserConsent

User = get_user_model()
F = TypeVar("F", bound=Callable[..., Any])


def track_activity(
    verb: str,
    get_actor: Optional[Callable] = None,
    get_action_object: Optional[Callable] = None,
    get_target: Optional[Callable] = None,
    get_data: Optional[Callable] = None,
) -> Callable[[F], F]:
    """
    Decorator to track activities in view functions.

    Args:
        verb: The verb describing the activity (e.g., 'viewed', 'downloaded')
        get_actor: Function to extract the actor from the view arguments
        get_action_object: Function to extract the action object from the view arguments
        get_target: Function to extract the target from the view arguments
        get_data: Function to extract additional data from the view arguments

    Returns:
        Decorated function that tracks the activity
    """

    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            # Extract request from args (typically the first or second argument in view functions)
            request = None
            for arg in list(args) + list(kwargs.values()):
                if isinstance(arg, HttpRequest):
                    request = arg
                    break

            # If no request was found, we can't check consent
            if not request:
                return func(*args, **kwargs)

            # Get the actor (default to the authenticated user)
            actor = None
            if get_actor:
                actor = get_actor(*args, **kwargs)
            elif hasattr(request, "user") and request.user.is_authenticated:
                actor = request.user

            # If no actor, we can't track the activity
            if not actor:
                return func(*args, **kwargs)

            # Get consent settings
            require_consent = getattr(settings, "ACTIVITY_CONSENT", {}).get(
                "REQUIRE_CONSENT", True
            )

            # If consent is not required, we can skip the consent check
            if not require_consent:
                # Just call the function and track the activity afterward
                result = func(*args, **kwargs)

                # Get the action object and target
                action_object = (
                    get_action_object(*args, **kwargs) if get_action_object else None
                )
                target = get_target(*args, **kwargs) if get_target else None

                # Get additional data
                data = get_data(*args, **kwargs) if get_data else {}

                # Record the activity
                record_activity(
                    actor=actor,
                    verb=verb,
                    action_object=action_object,
                    target=target,
                    request=request,
                    data=data,
                )

                return result

            # If consent is required, check if the user has given consent
            try:
                consent = UserConsent.objects.get(user=actor)
                if not consent.activity_tracking_enabled:
                    # User hasn't given consent, just call the function without tracking
                    return func(*args, **kwargs)
            except UserConsent.DoesNotExist:
                # No consent record, check default consent setting
                default_consent = getattr(settings, "ACTIVITY_CONSENT", {}).get(
                    "DEFAULT_CONSENT", False
                )
                if not default_consent:
                    # Default is no consent, just call the function without tracking
                    return func(*args, **kwargs)
                # Create a consent record with the default setting
                UserConsent.objects.get_or_create(
                    user=actor, defaults={"activity_tracking_enabled": default_consent}
                )

            # Get the action object and target
            action_object = (
                get_action_object(*args, **kwargs) if get_action_object else None
            )
            target = get_target(*args, **kwargs) if get_target else None

            # Get additional data
            data = get_data(*args, **kwargs) if get_data else {}

            # Call the function first
            result = func(*args, **kwargs)

            # Record the activity after the function has been called
            # This ensures we only track successful actions
            record_activity(
                actor=actor,
                verb=verb,
                action_object=action_object,
                target=target,
                request=request,
                data=data,
            )

            return result

        return cast(F, wrapper)

    return decorator



from typing import Any, Optional, Union, cast

from django.db.models import Q
from rest_framework import permissions
from strawberry.permission import BasePermission
from strawberry.types import Info

from api.models import Dataset, Organization
from authorization.models import DatasetPermission, OrganizationMembership, Role


# REST Framework Permissions
class IsOrganizationMember(permissions.BasePermission):
    """
    Permission class to check if a user is a member of the organization
    that owns the object being accessed.
    """

    def has_object_permission(self, request: Any, view: Any, obj: Any) -> bool:
        # If the user is a superuser, grant permission
        if request.user.is_superuser:
            return True

        # Get the organization from the object
        organization = self._get_organization(obj)
        if not organization:
            return False

        # Check if the user is a member of the organization
        return OrganizationMembership.objects.filter(
            user=request.user, organization=organization
        ).exists()

    def _get_organization(self, obj: Any) -> Optional[Organization]:
        """
        Get the organization from the object.
        Override this method in subclasses for specific object types.
        """
        if hasattr(obj, "organization"):
            return obj.organization  # type: ignore[attr-defined,no-any-return]
        return None


class HasOrganizationRole(permissions.BasePermission):
    """
    Permission class to check if a user has a specific role in the organization
    that owns the object being accessed.
    """

    def has_permission(self, request: Any, view: Any) -> bool:
        # For list views, check if user has access to any organization
        if request.user.is_superuser:
            return True

        # For organization-specific endpoints
        org_id = request.query_params.get("organization") or request.data.get(
            "organization"
        )
        if org_id:
            return OrganizationMembership.objects.filter(
                user=request.user, organization_id=org_id
            ).exists()  # type: ignore[no-any-return]

        # For general list views, allow if user belongs to any organization
        return OrganizationMembership.objects.filter(user=request.user).exists()

    def has_object_permission(self, request: Any, view: Any, obj: Any) -> bool:
        # If the user is a superuser, grant permission
        if request.user.is_superuser:
            return True

        # Get the organization from the object
        organization = self._get_organization(obj)
        if not organization:
            return False

        # Get the user's role in the organization
        try:
            membership = OrganizationMembership.objects.get(
                user=request.user, organization=organization
            )
            role = membership.role  # type: ignore[attr-defined]

            # Map the HTTP method to an action
            method = request.method.lower()
            action = self._get_action_from_method(method)

            # Check if the user's role allows the action
            if action == "view":
                return role.can_view  # type: ignore[attr-defined]
            elif action == "add":
                return role.can_add  # type: ignore[attr-defined]
            elif action == "change":
                return role.can_change  # type: ignore[attr-defined]
            elif action == "delete":
                return role.can_delete  # type: ignore[attr-defined]
            return False

        except OrganizationMembership.DoesNotExist:
            return False

    def _get_organization(self, obj: Any) -> Optional[Organization]:
        """
        Get the organization from the object.
        Override this method in subclasses for specific object types.
        """
        if hasattr(obj, "organization"):
            return obj.organization  # type: ignore[attr-defined,no-any-return]
        return None

    def _get_action_from_method(self, method: str) -> Optional[str]:
        """
        Map HTTP method to a permission action.
        """
        if method == "get":
            return "view"
        elif method == "post":
            return "add"
        elif method in ["put", "patch"]:
            return "change"
        elif method == "delete":
            return "delete"
        return None


# Strawberry GraphQL Permissions
class IsAuthenticated(BasePermission):
    """
    Permission class that checks if the user is authenticated.
    """

    message = "User is not authenticated"

    def has_permission(self, source: Any, info: Info, **kwargs: Any) -> bool:
        request = info.context
        return request.user.is_authenticated  # type: ignore[no-any-return]


class IsOrganizationMemberGraphQL(BasePermission):  # type: ignore[misc]
    """
    Permission class to check if a user is a member of the organization
    that owns the object being accessed.
    """

    message = "User is not a member of the organization"

    def has_permission(self, source: Any, info: Info, **kwargs: Any) -> bool:
        request = info.context

        # If the user is a superuser, grant permission
        if request.user.is_superuser:
            return True

        # For queries/mutations that don't have a source yet (e.g., creating a new object)
        if source is None:
            # Check if organization_id is provided in the arguments
            organization_id = kwargs.get("organization_id")
            if organization_id:
                return OrganizationMembership.objects.filter(
                    user=request.user, organization_id=organization_id
                ).exists()
            return True  # If no organization specified, allow and check later

        # For queries/mutations that have a source (e.g., updating an existing object)
        organization = self._get_organization(source)
        if not organization:
            return False

        return OrganizationMembership.objects.filter(
            user=request.user, organization=organization
        ).exists()

    def _get_organization(self, obj: Any) -> Optional[Organization]:
        """
        Get the organization from the object.
        Override this method in subclasses for specific object types.
        """
        if hasattr(obj, "organization"):
            return obj.organization  # type: ignore[attr-defined,no-any-return]
        return None


class HasOrganizationRoleGraphQL(BasePermission):  # type: ignore[misc]
    """
    Permission class to check if a user has a specific role in the organization
    that owns the object being accessed.
    """

    message = "User does not have the required role in the organization"

    def __init__(self, operation: str = "view"):
        self.operation = operation

    def has_permission(self, source: Any, info: Info, **kwargs: Any) -> bool:
        request = info.context

        # If the user is a superuser, grant permission
        if request.user.is_superuser:
            return True

        # For queries/mutations that don't have a source yet (e.g., creating a new object)
        if source is None:
            # Check if organization_id is provided in the arguments
            organization_id = kwargs.get("organization_id")
            # Also check if organization is in the context
            organization = None
            if hasattr(info.context, "context") and isinstance(
                info.context.context, dict
            ):
                organization = info.context.context.get("organization")

            if organization_id:
                try:
                    membership = OrganizationMembership.objects.get(
                        user=request.user, organization_id=organization_id
                    )
                    role = membership.role
                    return self._check_role_permission(role)  # type: ignore[no-any-return]
                except OrganizationMembership.DoesNotExist:
                    return False
            elif organization:
                try:
                    membership = OrganizationMembership.objects.get(
                        user=request.user, organization=organization
                    )
                    role = membership.role
                    return self._check_role_permission(role)  # type: ignore[no-any-return]
                except OrganizationMembership.DoesNotExist:
                    return False

            # If we're creating something that doesn't need organization permission yet,
            # we'll check later when the specific object is accessed
            return request.user.is_authenticated  # type: ignore[no-any-return]

        # For queries/mutations that have a source (e.g., updating an existing object)
        organization = self._get_organization(source)
        if not organization:
            return False

        try:
            membership = OrganizationMembership.objects.get(
                user=request.user, organization=organization
            )
            role = membership.role
            return self._check_role_permission(role)
        except OrganizationMembership.DoesNotExist:
            return False

    def _get_organization(self, obj: Any) -> Optional[Organization]:
        """
        Get the organization from the object.
        Override this method in subclasses for specific object types.
        """
        if hasattr(obj, "organization"):
            return obj.organization  # type: ignore[attr-defined,no-any-return]
        return None

    def _check_role_permission(self, role: Role) -> bool:
        """
        Check if the role has the required permission for the operation.
        """
        if self.operation == "view":
            return role.can_view
        elif self.operation == "add":
            return role.can_add
        elif self.operation == "change":
            return role.can_change
        elif self.operation == "delete":
            return role.can_delete
        return False


# Specialized permission classes for specific models
class DatasetPermissionGraphQL(HasOrganizationRoleGraphQL):  # type: ignore[misc]
    """
    Permission class specifically for Dataset objects.
    Also checks for dataset-specific permissions and user ownership.
    """

    def has_permission(self, source: Any, info: Info, **kwargs: Any) -> bool:
        # First check organization-level permissions
        if super().has_permission(source, info, **kwargs):
            return True

        # If not allowed at organization level, check dataset-specific permissions
        request = info.context

        # Check if user is authenticated before proceeding with permission checks
        if not hasattr(request, "user") or not request.user.is_authenticated:
            return False

        # For queries/mutations that don't have a source yet
        if source is None:
            dataset_id = kwargs.get("dataset_id")
            if not dataset_id:
                return False
            try:
                # Check if the user owns this dataset
                dataset = Dataset.objects.get(id=dataset_id)
                if dataset.user and dataset.user == request.user:
                    return True

                # If not owned by user, check dataset-specific permissions
                dataset_perm = DatasetPermission.objects.get(
                    user=request.user, dataset_id=dataset_id
                )
                role = dataset_perm.role
                return self._check_role_permission(role)
            except (Dataset.DoesNotExist, DatasetPermission.DoesNotExist):
                return False

        # For queries/mutations that have a source
        if hasattr(source, "id"):
            # Check if the user owns this dataset
            if hasattr(source, "user") and source.user == request.user:
                return True

            try:
                dataset_perm = DatasetPermission.objects.get(
                    user=request.user, dataset=source
                )
                role = dataset_perm.role
                return self._check_role_permission(role)
            except DatasetPermission.DoesNotExist:
                return False

        return False


class UserDatasetPermission(BasePermission):
    """
    Permission class that allows users to access their own datasets.
    """

    message = "You don't have permission to access this dataset"

    def has_permission(self, source: Any, info: Info, **kwargs: Any) -> bool:
        request = info.context

        # Check if user is authenticated
        if not hasattr(request, "user") or not request.user.is_authenticated:
            return False

        # For queries/mutations that don't have a source yet
        if source is None:
            dataset_id = kwargs.get("dataset_id")
            if dataset_id:
                try:
                    dataset = Dataset.objects.get(id=dataset_id)
                    # Allow access if user owns the dataset
                    return bool(dataset.user == request.user)
                except Dataset.DoesNotExist:
                    pass  # Let the resolver handle the non-existent dataset
            return False

        # For queries/mutations that have a source
        if hasattr(source, "user"):
            # Allow access if user owns the dataset
            return bool(source.user == request.user)

        return False


class CreateDatasetPermission(BasePermission):
    """
    Permission class for dataset creation.
    Allows users to create datasets either as part of an organization (if they have the right role)
    or as individual users.
    """

    message = "You don't have permission to create a dataset"

    def has_permission(self, source: Any, info: Info, **kwargs: Any) -> bool:
        request = info.context

        # Check if user is authenticated - basic requirement for all dataset creation
        if not hasattr(request, "user") or not request.user.is_authenticated:
            return False

        # If creating in organization context, check organization permissions
        organization = info.context.context.get("organization")
        if organization:
            try:
                # Check if user has the 'add' permission in the organization
                membership = OrganizationMembership.objects.get(
                    user=request.user, organization=organization
                )
                role = membership.role
                return role.can_add
            except OrganizationMembership.DoesNotExist:
                return False

        # If not in organization context, any authenticated user can create their own dataset
        return True


class ResourcePermissionGraphQL(HasOrganizationRoleGraphQL):  # type: ignore[misc]
    """
    Permission class specifically for Resource objects.
    """

    def _get_organization(self, obj: Any) -> Optional[Organization]:
        return obj.dataset.organization if hasattr(obj, "dataset") else None  # type: ignore[attr-defined,no-any-return]


class PublishDatasetPermission(BasePermission):
    """Permission class for publishing a dataset.
    Checks if the user has permission to publish the dataset.
    """

    message = "You don't have permission to publish/unpublish this dataset"

    def has_permission(self, source: Any, info: Info, **kwargs: Any) -> bool:
        request = info.context

        # Check if user is authenticated
        if not hasattr(request, "user") or not request.user.is_authenticated:
            return False

        # Superusers have access to everything
        if request.user.is_superuser:
            return True

        # Get the dataset ID from the arguments
        dataset_id = kwargs.get("dataset_id")
        if not dataset_id:
            return False

        try:
            dataset = Dataset.objects.get(id=dataset_id)

            # Check if user owns the dataset
            if dataset.user and dataset.user == request.user:
                return True

            # If organization-owned, check organization permissions
            if dataset.organization:
                # Get the roles with names 'admin', 'editor', or 'owner'
                admin_editor_roles = Role.objects.filter(
                    name__in=["admin", "editor", "owner"]
                ).values_list("id", flat=True)

                # Check if user is a member of the dataset's organization with appropriate role
                org_member = OrganizationMembership.objects.filter(
                    user=request.user,
                    organization=dataset.organization,
                    role__id__in=admin_editor_roles,
                ).exists()

                if org_member:
                    return True

            # Check dataset-specific permissions
            dataset_perm = DatasetPermission.objects.filter(
                user=request.user, dataset=dataset
            ).first()
            return bool(dataset_perm and dataset_perm.role.can_change)

        except Dataset.DoesNotExist:
            return False



import mimetypes
import os
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Optional, Union

import magic
import pandas as pd
import structlog

logger = structlog.get_logger(__name__)


def check_ext(file_object: Any) -> Optional[str]:
    ext_type = mimetypes.guess_type(file_object.name)[0]
    return ext_type


def check_mime_type(file_object: Any) -> Any:
    # with open(file_object.file) as mime_file:
    mime_type = magic.from_buffer(file_object.read(), mime=True)
    # file_object.file.close()
    return mime_type


def file_validation(
    file_object: Any, ext_object: Any, mapping: Dict[str, str]
) -> Optional[Any]:
    ext_type = check_ext(ext_object)
    print("ext--", ext_type)
    mime_type = check_mime_type(file_object)
    print("mime--", mime_type)
    if ext_type and mime_type:
        if mapping.get(ext_type.lower()) == mapping.get(mime_type.lower()):
            return mime_type
        else:
            return None
    else:
        return None


def get_file_extension(filename: str) -> Optional[str]:
    """Get the file extension from a filename."""
    _, ext = os.path.splitext(filename)
    return ext[1:] if ext else None


def get_mime_type(filename: str) -> Optional[str]:
    """Get the MIME type for a file."""
    mime_type, _ = mimetypes.guess_type(filename)
    return mime_type


def get_file_size(file_path: str) -> int:
    """Get the size of a file in bytes."""
    return os.path.getsize(file_path)


def ensure_directory_exists(directory_path: str) -> None:
    """Ensure a directory exists, create it if it doesn't."""
    os.makedirs(directory_path, exist_ok=True)


def get_relative_path(base_path: Union[str, Path], full_path: Union[str, Path]) -> str:
    """Get relative path from base path."""
    return os.path.relpath(full_path, base_path)


def join_paths(*paths: Union[str, Path]) -> str:
    """Join path components."""
    return os.path.join(*paths)


def normalize_path(path: Union[str, Path]) -> str:
    """Normalize path separators."""
    return os.path.normpath(str(path))


@lru_cache()
def load_csv(filepath: str) -> pd.DataFrame:
    """Load CSV file into a pandas DataFrame.

    Args:
        filepath: Path to the CSV file

    Returns:
        pd.DataFrame: Loaded DataFrame
    """
    return pd.read_csv(filepath)


@lru_cache()
def load_excel(filepath: str) -> pd.DataFrame:
    """Load Excel file into a pandas DataFrame.

    Args:
        filepath: Path to the Excel file

    Returns:
        pd.DataFrame: Loaded DataFrame
    """
    return pd.read_excel(filepath)


@lru_cache()
def load_ods(filepath: str) -> pd.DataFrame:
    """Load ODS file into a pandas DataFrame.

    Args:
        filepath: Path to the ODS file

    Returns:
        pd.DataFrame: Loaded DataFrame
    """
    return pd.read_excel(filepath, engine="odf")


@lru_cache()
def load_parquet(filepath: str) -> pd.DataFrame:
    """Load Parquet file into a pandas DataFrame.

    Args:
        filepath: Path to the Parquet file

    Returns:
        pd.DataFrame: Loaded DataFrame
    """
    return pd.read_parquet(filepath)


@lru_cache()
def load_feather(filepath: str) -> pd.DataFrame:
    """Load Feather file into a pandas DataFrame.

    Args:
        filepath: Path to the Feather file

    Returns:
        pd.DataFrame: Loaded DataFrame
    """
    return pd.read_feather(filepath)


@lru_cache()
def load_json(filepath: str) -> pd.DataFrame:
    """Load JSON file into a pandas DataFrame.

    Args:
        filepath: Path to the JSON file

    Returns:
        pd.DataFrame: Loaded DataFrame
    """
    return pd.read_json(filepath)


@lru_cache()
def load_tsv(filepath: str) -> pd.DataFrame:
    """Load TSV file into a pandas DataFrame.

    Args:
        filepath: Path to the TSV file

    Returns:
        pd.DataFrame: Loaded DataFrame
    """
    return pd.read_csv(filepath, sep="\t")


@lru_cache()
def load_tabular_data(filepath: str, format: str) -> Optional[pd.DataFrame]:
    """Load tabular data from various file formats into a pandas DataFrame.

    Args:
        filepath: Path to the file
        format: File format (csv, xls, xlsx, ods, parquet, feather, json, tsv)

    Returns:
        Optional[pd.DataFrame]: Loaded DataFrame or None if format not supported
    """
    format = format.lower()
    format_handlers = {
        "csv": load_csv,
        "xls": load_excel,
        "xlsx": load_excel,
        "ods": load_ods,
        "parquet": load_parquet,
        "feather": load_feather,
        "json": load_json,
        "tsv": load_tsv,
    }

    handler = format_handlers.get(format)
    if handler:
        try:
            return handler(filepath)
        except Exception as e:
            logger.error(f"Error loading {format} file: {str(e)}")
            return None
    return None



"""Base client for making API requests."""

from typing import Any, Dict, Optional

import requests

from dataspace_sdk.exceptions import (
    DataSpaceAPIError,
    DataSpaceAuthError,
    DataSpaceNotFoundError,
    DataSpaceValidationError,
)


class BaseAPIClient:
    """Base client for making API requests to DataSpace."""

    def __init__(self, base_url: str, auth_client: Any = None):
        """
        Initialize the base API client.

        Args:
            base_url: Base URL of the DataSpace API
            auth_client: Authentication client instance
        """
        self.base_url = base_url.rstrip("/")
        self.auth_client = auth_client

    def _get_headers(self, additional_headers: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        """
        Get request headers including authentication.

        Args:
            additional_headers: Additional headers to include

        Returns:
            Dictionary of headers
        """
        headers = {"Content-Type": "application/json"}

        if self.auth_client and self.auth_client.is_authenticated():
            headers["Authorization"] = f"Bearer {self.auth_client.access_token}"

        if additional_headers:
            headers.update(additional_headers)

        return headers

    def _make_request(
        self,
        method: str,
        endpoint: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        json_data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> Dict[str, Any]:
        """
        Make an HTTP request to the API.

        Args:
            method: HTTP method (GET, POST, etc.)
            endpoint: API endpoint
            params: Query parameters
            data: Form data
            json_data: JSON data
            headers: Additional headers

        Returns:
            Response data as dictionary

        Raises:
            DataSpaceAPIError: For API errors
            DataSpaceAuthError: For authentication errors
            DataSpaceNotFoundError: For 404 errors
            DataSpaceValidationError: For validation errors
        """
        url = f"{self.base_url}{endpoint}"
        request_headers = self._get_headers(headers)

        try:
            response = requests.request(
                method=method,
                url=url,
                params=params,
                data=data,
                json=json_data,
                headers=request_headers,
            )

            # Handle different status codes
            if response.status_code == 200 or response.status_code == 201:
                result: Dict[str, Any] = response.json() if response.content else {}
                return result
            elif response.status_code == 204:
                return {}
            elif response.status_code == 401:
                raise DataSpaceAuthError(
                    "Authentication required or token expired",
                    status_code=response.status_code,
                    response=response.json() if response.content else {},
                )
            elif response.status_code == 404:
                raise DataSpaceNotFoundError(
                    "Resource not found",
                    status_code=response.status_code,
                    response=response.json() if response.content else {},
                )
            elif response.status_code == 400:
                raise DataSpaceValidationError(
                    "Validation error",
                    status_code=response.status_code,
                    response=response.json() if response.content else {},
                )
            else:
                raise DataSpaceAPIError(
                    f"API request failed with status {response.status_code}",
                    status_code=response.status_code,
                    response=response.json() if response.content else {},
                )

        except requests.RequestException as e:
            raise DataSpaceAPIError(f"Network error: {str(e)}")

    def get(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        """Make a GET request."""
        return self._make_request("GET", endpoint, params=params)

    def post(
        self,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        json_data: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Make a POST request."""
        return self._make_request("POST", endpoint, data=data, json_data=json_data)

    def put(
        self,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        json_data: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Make a PUT request."""
        return self._make_request("PUT", endpoint, data=data, json_data=json_data)

    def patch(
        self,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        json_data: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Make a PATCH request."""
        return self._make_request("PATCH", endpoint, data=data, json_data=json_data)

    def delete(self, endpoint: str) -> Dict[str, Any]:
        """Make a DELETE request."""
        response = self._make_request("DELETE", endpoint)
        result: Dict[str, Any] = response if isinstance(response, dict) else {}
        return result



from typing import Any, Dict, List, Optional, Union, cast

from django.db import DataError, IntegrityError
from graphql import GraphQLError
from strawberry.extensions import Extension
from strawberry.types import ExecutionContext

from api.utils.error_handlers import (
    ErrorDictType,
    FieldErrors,
    NonFieldErrors,
    format_data_error,
    format_integrity_error,
    is_field_errors,
    is_non_field_errors,
)


class ErrorFormatterExtension(Extension):  # type: ignore[misc,valid-type]
    def process_errors(
        self, errors: List[GraphQLError], execution_context: ExecutionContext
    ) -> List[GraphQLError]:
        formatted_errors = []
        for error in errors:
            original = getattr(error, "original_error", error)

            if isinstance(original, (DataError, IntegrityError)):
                error_data = (
                    format_data_error(original)
                    if isinstance(original, DataError)
                    else format_integrity_error(original)
                )
                if is_field_errors(error_data):
                    field_errors = error_data["field_errors"]
                    if not field_errors:
                        formatted_errors.append(error)
                        continue
                    # Get the first error message safely
                    first_field = next(iter(field_errors.keys()))
                    first_message = (
                        field_errors[first_field][0]
                        if field_errors[first_field]
                        else "Field validation error"
                    )
                    formatted_errors.append(
                        GraphQLError(
                            message=first_message,
                            path=error.path,
                            extensions={"field_errors": field_errors},
                        )
                    )
                elif is_non_field_errors(error_data):
                    non_field_errors = error_data["non_field_errors"]
                    if not non_field_errors:
                        formatted_errors.append(error)
                        continue
                    # Get the first error message safely
                    first_message = (
                        non_field_errors[0]
                        if non_field_errors
                        else "Non-field validation error"
                    )
                    formatted_errors.append(
                        GraphQLError(
                            message=first_message,
                            path=error.path,
                            extensions={"non_field_errors": non_field_errors},
                        )
                    )
            else:
                formatted_errors.append(error)

        return formatted_errors



# Generated by Django 5.0.4 on 2024-10-15 07:22

import api.utils.file_paths
import django.db.models.deletion
import uuid
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Metadata',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.CharField(max_length=75)),
                ('data_standard', models.CharField(blank=True, choices=[('DCATV3', 'Dcatv3'), ('OCDS', 'Ocds'), ('OBDS', 'Obds'), ('NA', 'Na')], max_length=50)),
                ('urn', models.CharField(blank=True, max_length=175, unique=True)),
                ('data_type', models.CharField(choices=[('STRING', 'String'), ('NUMBER', 'Number'), ('SELECT', 'Select'), ('MULTISELECT', 'Multiselect'), ('DATE', 'Date'), ('URL', 'Url')], max_length=50)),
                ('options', models.JSONField(blank=True, null=True)),
                ('validator', models.JSONField(blank=True, default=list)),
                ('validator_options', models.JSONField(blank=True, null=True)),
                ('type', models.CharField(choices=[('REQUIRED', 'Required'), ('OPTIONAL', 'Optional'), ('ADVANCED', 'Advanced')], max_length=50)),
                ('model', models.CharField(choices=[('DATASET', 'Dataset'), ('RESOURCE', 'Reseource')], max_length=50)),
                ('enabled', models.BooleanField(default=False)),
                ('filterable', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='ResourcePreviewDetails',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_all_entries', models.BooleanField(default=True)),
                ('start_entry', models.IntegerField(default=0)),
                ('end_entry', models.IntegerField(default=10)),
            ],
        ),
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(max_length=50, unique=True)),
            ],
            options={
                'verbose_name': 'Tag',
                'verbose_name_plural': 'Tags',
            },
        ),
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=75, unique=True)),
                ('description', models.CharField(blank=True, max_length=1000, null=True)),
                ('slug', models.SlugField(max_length=75, null=True, unique=True)),
                ('parent_id', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='api.category')),
            ],
        ),
        migrations.CreateModel(
            name='Dataset',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('title', models.CharField(blank=True, max_length=300)),
                ('description', models.CharField(blank=True, max_length=1000)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('status', models.CharField(choices=[('DRAFT', 'Draft'), ('PUBLISHED', 'Published'), ('ARCHIVED', 'Archived')], default='DRAFT', max_length=50)),
                ('categories', models.ManyToManyField(blank=True, to='api.category')),
            ],
        ),
        migrations.CreateModel(
            name='Geography',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=75, unique=True)),
                ('code', models.CharField(blank=True, default='', max_length=100, null=True)),
                ('type', models.CharField(choices=[('DISTRICT', 'District'), ('STATE', 'State'), ('COUNTRY', 'Country'), ('UT', 'Ut')], max_length=20)),
                ('parent_id', models.ForeignKey(blank=True, default='', null=True, on_delete=django.db.models.deletion.CASCADE, to='api.geography')),
            ],
        ),
        migrations.CreateModel(
            name='DatasetMetadata',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(max_length=1000)),
                ('dataset', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='metadata', to='api.dataset')),
                ('metadata_item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.metadata')),
            ],
        ),
        migrations.CreateModel(
            name='Organization',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
                ('description', models.CharField(max_length=1000)),
                ('logo', models.ImageField(blank=True, null=True, upload_to=api.utils.file_paths._organization_directory_path)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('homepage', models.URLField(blank=True)),
                ('contact_email', models.EmailField(blank=True, max_length=254, null=True)),
                ('organization_types', models.CharField(choices=[('STATE GOVERNMENT', 'State Government'), ('UNION TERRITORY GOVERNMENT', 'Union Territory Government'), ('URBAN LOCAL BODY', 'Urban Local Body'), ('ACADEMIC INSTITUTION', 'Academic Institution'), ('CENTRAL GOVERNMENT', 'Central Government'), ('CITIZENS GROUP', 'Citizens Group'), ('CIVIL SOCIETY ORGANISATION', 'Civil Society Organisation'), ('INDUSTRY BODY', 'Industry Body'), ('MEDIA ORGANISATION', 'Media Organisation'), ('OPEN DATA/TECHNOLOGY COMMUNITY', 'Open Data Technology Community'), ('PRIVATE COMPANY', 'Private Company'), ('PUBLIC SECTOR COMPANY', 'Public Sector Company'), ('OTHERS', 'Others'), ('STARTUP', 'Startup'), ('GOVERNMENT', 'Government'), ('CORPORATIONS', 'Corporations'), ('NGO', 'Ngo')], max_length=50)),
                ('slug', models.SlugField(max_length=75, null=True, unique=True)),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='parent_field', to='api.organization')),
            ],
        ),
        migrations.AddField(
            model_name='dataset',
            name='organization',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='api.organization'),
        ),
        migrations.CreateModel(
            name='AccessModel',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(blank=True, max_length=300, null=True)),
                ('description', models.CharField(blank=True, max_length=1000, null=True)),
                ('type', models.CharField(choices=[('PUBLIC', 'Public'), ('PRIVATE', 'Private'), ('PROTECTED', 'Protected')], default='PUBLIC', max_length=100)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('dataset', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.dataset')),
                ('organization', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='api.organization')),
            ],
        ),
        migrations.CreateModel(
            name='Resource',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('type', models.CharField(choices=[('API', 'Api'), ('FILE', 'File'), ('EXTERNAL', 'External')], default='FILE', max_length=50)),
                ('name', models.CharField(blank=True, max_length=50)),
                ('description', models.CharField(blank=True, max_length=1000)),
                ('preview_enabled', models.BooleanField(default=False)),
                ('dataset', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='resources', to='api.dataset')),
                ('preview_details', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='api.resourcepreviewdetails')),
            ],
        ),
        migrations.CreateModel(
            name='ResourceFileDetails',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('file', models.FileField(upload_to='resources/')),
                ('size', models.FloatField(blank=True, null=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('format', models.CharField(max_length=50)),
                ('resource', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='api.resource')),
            ],
        ),
        migrations.CreateModel(
            name='ResourceMetadata',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(max_length=1000)),
                ('metadata_item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.metadata')),
                ('resource', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.resource')),
            ],
        ),
        migrations.CreateModel(
            name='ResourceSchema',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('field_name', models.CharField(max_length=255)),
                ('format', models.CharField(choices=[('STRING', 'String'), ('NUMBER', 'Number'), ('INTEGER', 'Integer'), ('DATE', 'Date')], max_length=255)),
                ('description', models.CharField(blank=True, max_length=1000, null=True)),
                ('resource', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.resource')),
            ],
        ),
        migrations.CreateModel(
            name='ResourceChartDetails',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(blank=True, max_length=50)),
                ('description', models.CharField(blank=True, max_length=1000)),
                ('chart_type', models.CharField(choices=[('BAR_VERTICAL', 'Bar Vertical'), ('BAR_HORIZONTAL', 'Bar Horizontal'), ('LINE', 'Line'), ('ASSAM_DISTRICT', 'Assam District'), ('ASSAM_RC', 'Assam Rc')], default='BAR_HORIZONTAL', max_length=50)),
                ('x_axis_label', models.CharField(default='', max_length=50)),
                ('y_axis_label', models.CharField(default='', max_length=50)),
                ('show_legend', models.BooleanField(default=False)),
                ('aggregate_type', models.CharField(choices=[('NONE', 'None'), ('SUM', 'Sum'), ('AVERAGE', 'Average'), ('COUNT', 'Count')], default='NONE', max_length=50)),
                ('resource', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.resource')),
                ('region_column', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='region', to='api.resourceschema')),
                ('value_column', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='value', to='api.resourceschema')),
                ('x_axis_column', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='x_column', to='api.resourceschema')),
                ('y_axis_column', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='y_column', to='api.resourceschema')),
            ],
        ),
        migrations.CreateModel(
            name='AccessModelResource',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('access_model', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.accessmodel')),
                ('resource', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='api.resource')),
                ('fields', models.ManyToManyField(to='api.resourceschema')),
            ],
        ),
        migrations.AddField(
            model_name='dataset',
            name='tags',
            field=models.ManyToManyField(blank=True, to='api.tag'),
        ),
    ]



"""
Test settings for Django tests.
"""

import os
import sys

# Add the project root directory to Python path
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, project_root)

from DataSpace.settings import *

# Use an in-memory SQLite database for testing
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

# Disable debug mode
DEBUG = True

# Use a faster password hasher during tests
PASSWORD_HASHERS = [
    'django.contrib.auth.hashers.MD5PasswordHasher',
]

# Disable migrations during tests
class DisableMigrations:
    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        return None

MIGRATION_MODULES = DisableMigrations()

# Disable celery tasks during tests
CELERY_ALWAYS_EAGER = True
CELERY_EAGER_PROPAGATES_EXCEPTIONS = True

# Required apps for testing
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'api',
]



"""Custom exceptions for DataSpace SDK."""

from typing import Any, Optional


class DataSpaceAPIError(Exception):
    """Base exception for DataSpace API errors."""

    def __init__(
        self,
        message: str,
        status_code: Optional[int] = None,
        response: Optional[dict[str, Any]] = None,
    ):
        self.message = message
        self.status_code = status_code
        self.response = response
        super().__init__(self.message)


class DataSpaceAuthError(DataSpaceAPIError):
    """Exception raised for authentication errors."""

    pass


class DataSpaceNotFoundError(DataSpaceAPIError):
    """Exception raised when a resource is not found."""

    pass


class DataSpaceValidationError(DataSpaceAPIError):
    """Exception raised for validation errors."""

    pass



# mypy: disable-error-code=no-untyped-def
from django.apps import AppConfig


class AuthorizationConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "authorization"

    def ready(self):
        # Import and register models with activity stream
        try:
            from authorization.activity_registry import register_models

            register_models()
        except ImportError:
            # This prevents issues during migrations when models might not be available yet
            pass



from typing import Any

import structlog
from django.core.cache import cache
from django.db.models.signals import pre_save
from django.dispatch import receiver

from api.models.UseCase import UseCase
from api.utils.enums import UseCaseStatus
from search.documents.usecase_document import UseCaseDocument

# Cache version key for search results
SEARCH_CACHE_VERSION_KEY = "search_results_version"

logger = structlog.getLogger(__name__)


@receiver(pre_save, sender=UseCase)
def handle_usecase_publication(sender: Any, instance: UseCase, **kwargs: Any) -> None:
    """
    Signal handler for usecase publication state changes.

    This will:
    1. Update the Elasticsearch index when status changes to/from PUBLISHED
    2. Invalidate search cache when status changes
    """
    try:
        # Check if this is an existing usecase (not a new one)
        if instance.pk:
            # Get the original usecase from the database
            original = UseCase.objects.get(pk=instance.pk)

            # Check if status is changing to/from PUBLISHED
            status_changing_to_published = (
                original.status != UseCaseStatus.PUBLISHED
                and instance.status == UseCaseStatus.PUBLISHED
            )
            status_changing_from_published = (
                original.status == UseCaseStatus.PUBLISHED
                and instance.status != UseCaseStatus.PUBLISHED
            )

            # Only proceed if status is actually changing
            if status_changing_to_published or status_changing_from_published:
                # Invalidate search results cache by incrementing version
                try:
                    version = cache.get(SEARCH_CACHE_VERSION_KEY, 0)
                    cache.set(SEARCH_CACHE_VERSION_KEY, version + 1)
                    logger.info(
                        f"Invalidated search cache for usecase {instance.title}"
                    )
                except Exception as e:
                    logger.error(f"Failed to invalidate search cache: {str(e)}")

                # Update Elasticsearch index
                if status_changing_from_published:
                    # Remove from index when unpublished
                    try:
                        document = UseCaseDocument.get(id=instance.id, ignore=404)
                        if document:
                            document.delete()
                            logger.info(
                                f"Removed usecase {instance.title} from Elasticsearch index"
                            )
                    except Exception as e:
                        logger.error(
                            f"Failed to delete Elasticsearch document for usecase {instance.title}: {str(e)}"
                        )
                elif status_changing_to_published:
                    # Add to index when published
                    try:
                        document = UseCaseDocument()
                        document.update(instance)
                        logger.info(
                            f"Added usecase {instance.title} to Elasticsearch index"
                        )
                    except Exception as e:
                        logger.error(
                            f"Failed to add Elasticsearch document for usecase {instance.title}: {str(e)}"
                        )

    except Exception as e:
        logger.error(f"Error in usecase publication signal handler: {str(e)}")
        # Don't raise the exception to avoid blocking the save operation



from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Protocol, cast

from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from django.db import models

from api.utils.enums import (
    MetadataDataTypes,
    MetadataModels,
    MetadataStandards,
    MetadataTypes,
)
from api.utils.metadata_validators import VALIDATOR_MAP


class MetadataType(Protocol):
    id: int
    label: str
    data_type: str
    type: str
    options: Optional[List[str]]
    validator: List[str]
    validator_options: Any


class BaseMetadata(models.Model):
    """Base class for all metadata models."""

    id = models.AutoField(primary_key=True)
    metadata_item = models.ForeignKey(
        "api.Metadata", on_delete=models.CASCADE, null=False, blank=False
    )
    value = models.CharField(max_length=1000, unique=False)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True
        ordering = ["metadata_item__label"]

    def clean(self) -> None:
        """
        Custom validation logic to validate the value against metadata_item's options.
        """
        metadata = cast(MetadataType, self.metadata_item)
        value = str(self.value)
        self._validate_data_type(metadata, value)
        self._apply_custom_validators(metadata, value)

    def _apply_custom_validators(self, metadata: MetadataType, value: str) -> None:
        """
        Apply user-selected custom validators.
        """
        selected_validators: List[str] = getattr(metadata, "validator", [])

        for validator_name in selected_validators:
            validator = VALIDATOR_MAP.get(validator_name)

            if validator:
                if validator_name == "regex_validator":
                    pattern = getattr(metadata, "validator_options", "")
                    validator(value, pattern)  # type: ignore
                else:
                    validator(value)  # type: ignore
            else:
                raise ValidationError(f"Unknown validator: {validator_name}")

    def _validate_data_type(self, metadata: MetadataType, value: str) -> None:
        validation_methods: Dict[str, Callable[[MetadataType, str], None]] = {
            MetadataDataTypes.STRING: self._validate_string,
            MetadataDataTypes.NUMBER: self._validate_number,
            MetadataDataTypes.DATE: self._validate_date,
            MetadataDataTypes.URL: self._validate_url,
            MetadataDataTypes.SELECT: self._validate_select,
            MetadataDataTypes.MULTISELECT: self._validate_multiselect,
        }
        # Get the corresponding validation method based on the data_type
        data_type: Optional[str] = getattr(metadata, "data_type", None)
        validate_method = validation_methods.get(data_type) if data_type else None

        metadata_type = getattr(metadata, "type", None)
        if not value and metadata_type is MetadataTypes.REQUIRED:
            raise ValidationError(
                f"Required value not sent for: {getattr(metadata, 'label', '')}"
            )

        if validate_method:
            try:
                validate_method(metadata, value)
            except ValidationError as e:
                # if metadata_type is MetadataTypes.REQUIRED:
                raise e
                # else:
                #     # Set empty value for non-required fields with validation errors
                #     self.value = ""
        else:
            raise ValidationError(f"Unknown data type: {data_type}")

    def _validate_string(self, metadata: MetadataType, value: str) -> None:
        """Validate string type."""
        if not isinstance(value, str):
            raise ValidationError(
                f"Value for '{getattr(metadata, 'label', '')}' must be a string."
            )

    def _validate_number(self, metadata: MetadataType, value: str) -> None:
        """Validate number type."""
        try:
            float(value)
        except ValueError:
            raise ValidationError(
                f"Value for '{getattr(metadata, 'label', '')}' must be a valid number."
            )

    def _validate_select(self, metadata: MetadataType, value: str) -> None:
        """Validate singleselect type."""
        options = getattr(metadata, "options", [])
        if value not in options:
            raise ValidationError(
                f"Invalid value: '{value}' for '{getattr(metadata, 'label', '')}'. Must be one of {options}."
            )

    def _validate_multiselect(self, metadata: MetadataType, value: str) -> None:
        """Validate multiselect type."""
        options = getattr(metadata, "options", [])
        selected_values = [v.strip() for v in value.split(",")]
        invalid_values = [v for v in selected_values if v not in options]
        if invalid_values:
            raise ValidationError(
                f"Invalid values: {', '.join(invalid_values)} for '{getattr(metadata, 'label', '')}'. Must be one of {options}."
            )

    def _validate_date(self, metadata: MetadataType, value: str) -> None:
        """Validate date type."""
        try:
            datetime.strptime(value, "%Y-%m-%d")
        except ValueError:
            raise ValidationError(
                f"Value for '{getattr(metadata, 'label', '')}' must be a valid date in YYYY-MM-DD format."
            )

    def _validate_url(self, metadata: MetadataType, value: str) -> None:
        """Validate URL type."""
        validator = URLValidator()
        try:
            validator(value)
        except ValidationError:
            raise ValidationError(
                f"Value for '{getattr(metadata, 'label', '')}' must be a valid URL."
            )

    def save(self, *args: Any, **kwargs: Any) -> None:
        """
        Override save to run validation before saving.
        """
        self.clean()
        super().save(*args, **kwargs)


class Metadata(models.Model):
    id = models.AutoField(primary_key=True)
    label = models.CharField(max_length=75, unique=False)
    data_standard = models.CharField(
        max_length=50,
        choices=MetadataStandards.choices,
        blank=True,
        unique=False,
        null=True,
    )
    urn = models.CharField(max_length=175, unique=True, blank=True)
    data_type = models.CharField(
        max_length=50, choices=MetadataDataTypes.choices, blank=False, unique=False
    )
    options = models.JSONField(blank=True, null=True)  # for select and multiselect
    validator = models.JSONField(blank=True, default=list)  # predefined set
    validator_options = models.JSONField(
        blank=True, null=True
    )  # options for validation
    type = models.CharField(
        max_length=50, choices=MetadataTypes.choices, blank=False, unique=False
    )
    model = models.CharField(
        max_length=50, choices=MetadataModels.choices, blank=False, unique=False
    )
    enabled = models.BooleanField(default=False)
    filterable = models.BooleanField(default=False)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    def __str__(self) -> str:
        return f"{self.label} ({self.data_type})"

    class Meta:
        db_table = "metadata"
        ordering = ["label"]



"""Telemetry utilities for tracing and metrics."""

import functools
import time
from typing import Any, Callable, Dict, Optional, TypeVar, cast

from opentelemetry import metrics, trace
from opentelemetry.metrics import Counter, Histogram, UpDownCounter
from opentelemetry.trace import Span, Status, StatusCode

# Type variable for generic function type
F = TypeVar("F", bound=Callable[..., Any])

# Get tracer and meter
tracer = trace.get_tracer(__name__)
meter = metrics.get_meter(__name__)

# Define metrics
http_request_duration = meter.create_histogram(
    name="http_request_duration",
    description="Duration of HTTP requests",
    unit="ms",
)

api_requests_total = meter.create_counter(
    name="api_requests_total",
    description="Total number of API requests",
)

active_requests = meter.create_up_down_counter(
    name="active_requests",
    description="Number of currently active requests",
)


def trace_method(
    name: Optional[str] = None,
    attributes: Optional[Dict[str, str]] = None,
    record_exceptions: bool = True,
) -> Callable[[F], F]:
    """Decorator to add OpenTelemetry tracing to a method.

    Args:
        name: Optional name for the span. If not provided, uses the function name.
        attributes: Optional attributes to add to the span.
        record_exceptions: Whether to record exceptions in the span.
    """

    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            # Get span name from parameter or function name
            span_name = name or func.__name__

            with tracer.start_as_current_span(span_name) as span:
                # Add attributes
                if attributes:
                    for key, value in attributes.items():
                        span.set_attribute(key, value)

                # Add function parameters as span attributes
                try:
                    # Skip 'self' parameter for instance methods
                    params = dict(zip(func.__code__.co_varnames[1:], args[1:]))
                    params.update(kwargs)
                    for key, value in params.items():
                        # Only add simple types as attributes
                        if isinstance(value, (str, int, float, bool)):
                            span.set_attribute(f"parameter.{key}", str(value))
                except Exception:
                    pass  # Skip parameter recording if it fails

                try:
                    # Record start time for metrics
                    start_time = time.time()

                    # Execute the function
                    result = func(*args, **kwargs)

                    # Record duration
                    duration_ms = (time.time() - start_time) * 1000
                    http_request_duration.record(duration_ms)

                    # Set span status
                    span.set_status(Status(StatusCode.OK))

                    return result

                except Exception as e:
                    if record_exceptions:
                        # Record exception in span
                        span.record_exception(e)
                        span.set_status(Status(StatusCode.ERROR, str(e)))
                    raise

        return cast(F, wrapper)

    return decorator


def track_metrics(
    name: Optional[str] = None,
    record_duration: bool = True,
) -> Callable[[F], F]:
    """Decorator to record metrics for a function.

    Args:
        name: Optional name prefix for metrics. If not provided, uses the function name.
        record_duration: Whether to record duration metrics.
    """

    def decorator(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            metric_name = name or func.__name__

            # Increment request counter
            api_requests_total.add(1, {"endpoint": metric_name})

            # Track active requests
            active_requests.add(1, {"endpoint": metric_name})

            try:
                start_time = time.time()
                result = func(*args, **kwargs)

                if record_duration:
                    duration_ms = (time.time() - start_time) * 1000
                    http_request_duration.record(duration_ms, {"endpoint": metric_name})

                return result

            finally:
                # Decrement active requests
                active_requests.add(-1, {"endpoint": metric_name})

        return cast(F, wrapper)

    return decorator



"""DataSpace Python SDK for programmatic access to DataSpace resources."""

from dataspace_sdk.client import DataSpaceClient
from dataspace_sdk.exceptions import (
    DataSpaceAPIError,
    DataSpaceAuthError,
    DataSpaceNotFoundError,
    DataSpaceValidationError,
)

__version__ = "0.1.0"
__all__ = [
    "DataSpaceClient",
    "DataSpaceAPIError",
    "DataSpaceAuthError",
    "DataSpaceNotFoundError",
    "DataSpaceValidationError",
]



import unittest
import uuid
from typing import List, Optional
from unittest.mock import Mock, patch

from api.models import Dataset, Metadata
from api.models.DatasetMetadata import DatasetMetadata
from api.schema.dataset_schema import DSMetadataItemType, _add_update_dataset_metadata


class TestAddUpdateDatasetMetadata(unittest.TestCase):
    """Test cases for _add_update_dataset_metadata function."""

    def setUp(self) -> None:
        """Set up test fixtures."""
        self.dataset_id = uuid.uuid4()
        self.dataset: Mock = Mock(spec=Dataset, id=self.dataset_id)

    @patch("api.schema.dataset_schema.Dataset.objects.create")
    @patch("api.schema.dataset_schema.Metadata.objects.get")
    def test_add_new_metadata(
        self, mock_metadata_get: Mock, mock_dataset_create: Mock
    ) -> None:
        """Test adding new metadata to a dataset."""
        mock_dataset_create.return_value = self.dataset
        mock_metadata_get.return_value = Mock(spec=Metadata, enabled=True)

        metadata_input: List[DSMetadataItemType] = [
            DSMetadataItemType(id="1", value="Value 1"),
            DSMetadataItemType(id="2", value="Value 2"),
        ]

        with patch(
            "api.schema.dataset_schema._delete_existing_metadata"
        ) as mock_delete:
            _add_update_dataset_metadata(self.dataset, metadata_input)
            mock_delete.assert_called_once_with(self.dataset)

        mock_dataset_create.assert_called_once()
        mock_metadata_get.assert_any_call(id="1")
        mock_metadata_get.assert_any_call(id="2")

        self.assertEqual(self.dataset.datasetmetadata_set.count(), 2)

    @patch("api.schema.dataset_schema.Dataset.objects.create")
    @patch("api.schema.dataset_schema.Metadata.objects.get")
    def test_update_existing_metadata(
        self, mock_metadata_get: Mock, mock_dataset_create: Mock
    ) -> None:
        """Test updating existing metadata in a dataset."""
        mock_dataset_create.return_value = self.dataset
        mock_metadata_get.return_value = Mock(spec=Metadata, enabled=True)

        metadata_input: List[DSMetadataItemType] = [
            DSMetadataItemType(id="1", value="Value 1"),
            DSMetadataItemType(id="2", value="Value 2"),
        ]
        _add_update_dataset_metadata(self.dataset, metadata_input)

        updated_metadata_input: List[DSMetadataItemType] = [
            DSMetadataItemType(id="1", value="Updated Value 1"),
            DSMetadataItemType(id="2", value="Updated Value 2"),
        ]
        _add_update_dataset_metadata(self.dataset, updated_metadata_input)

        mock_dataset_create.assert_called_once()
        mock_metadata_get.assert_any_call(id="1")
        mock_metadata_get.assert_any_call(id="2")

        self.assertEqual(self.dataset.datasetmetadata_set.count(), 2)

    @patch("api.schema.dataset_schema.Dataset.objects.create")
    @patch("api.schema.dataset_schema.Metadata.objects.get")
    def test_delete_existing_metadata(
        self, mock_metadata_get: Mock, mock_dataset_create: Mock
    ) -> None:
        """Test deleting existing metadata from a dataset."""
        mock_dataset_create.return_value = self.dataset
        mock_metadata_get.return_value = Mock(spec=Metadata, enabled=True)

        metadata_input: List[DSMetadataItemType] = [
            DSMetadataItemType(id="1", value="Value 1"),
            DSMetadataItemType(id="2", value="Value 2"),
        ]
        _add_update_dataset_metadata(self.dataset, metadata_input)
        self.assertEqual(self.dataset.datasetmetadata_set.count(), 2)

        with patch(
            "api.schema.dataset_schema._delete_existing_metadata"
        ) as mock_delete:
            _add_update_dataset_metadata(self.dataset, [])
            mock_delete.assert_called_once_with(self.dataset)

        self.assertEqual(self.dataset.datasetmetadata_set.count(), 0)

    @patch("api.schema.dataset_schema.Dataset.objects.create")
    @patch("api.schema.dataset_schema.Metadata.objects.get")
    def test_handle_non_existent_metadata_id(
        self, mock_metadata_get: Mock, mock_dataset_create: Mock
    ) -> None:
        """Test handling non-existent metadata ID."""
        mock_dataset_create.return_value = self.dataset
        mock_metadata_get.side_effect = Metadata.DoesNotExist

        metadata_input: List[DSMetadataItemType] = [
            DSMetadataItemType(id="non-existent-id", value="Value 1"),
        ]

        with self.assertRaises(ValueError) as context:
            _add_update_dataset_metadata(self.dataset, metadata_input)

        self.assertEqual(
            str(context.exception), "Metadata with ID non-existent-id does not exist."
        )

    @patch("api.schema.dataset_schema.Dataset.objects.create")
    @patch("api.schema.dataset_schema.Metadata.objects.get")
    def test_handle_disabled_metadata_field(
        self, mock_metadata_get: Mock, mock_dataset_create: Mock
    ) -> None:
        """Test handling disabled metadata field."""
        mock_dataset_create.return_value = self.dataset
        mock_metadata_get.return_value = Mock(spec=Metadata, enabled=False)

        metadata_input: List[DSMetadataItemType] = [
            DSMetadataItemType(id="1", value="Value 1"),
        ]

        with self.assertRaises(ValueError) as context:
            _add_update_dataset_metadata(self.dataset, metadata_input)

        self.assertEqual(str(context.exception), "Metadata with ID 1 is not enabled.")



"""Example: Working with organization resources."""

from typing import Any, List

from dataspace_sdk import DataSpaceClient

# Initialize and login
client = DataSpaceClient(base_url="https://api.dataspace.example.com")
keycloak_token = "your_keycloak_token_here"

try:
    user_info = client.login(keycloak_token=keycloak_token)
    print(f"Logged in as: {user_info['user']['username']}\n")

    # Iterate through user's organizations
    for org in user_info["user"]["organizations"]:
        print(f"Organization: {org['name']}")
        print(f"Role: {org['role']}")
        print("-" * 50)

        # Get datasets for this organization
        datasets = client.datasets.get_organization_datasets(organization_id=org["id"], limit=10)
        # Handle both list and dict responses
        datasets_list: List[Any] = datasets if isinstance(datasets, list) else []
        print(f"Datasets: {len(datasets_list)}")
        for dataset in datasets_list[:5]:
            print(f"  - {dataset['title']}")

        # Get AI models for this organization
        models = client.aimodels.get_organization_models(organization_id=org["id"], limit=10)
        # Handle both list and dict responses
        models_list: List[Any] = models if isinstance(models, list) else []
        print(f"\nAI Models: {len(models_list)}")
        for model in models_list[:5]:
            print(f"  - {model.get('displayName', model.get('name'))}")

        # Get use cases for this organization
        usecases = client.usecases.get_organization_usecases(organization_id=org["id"], limit=10)
        # Handle both list and dict responses
        usecases_list: List[Any] = usecases if isinstance(usecases, list) else []
        print(f"\nUse Cases: {len(usecases_list)}")
        for usecase in usecases_list[:5]:
            print(f"  - {usecase['title']}")

        print("\n")

except Exception as e:
    print(f"Error: {e}")



import time
import uuid
from typing import Any, Dict, List, Optional, cast

import structlog
from django.core.management.base import BaseCommand, CommandParser
from django.db import transaction

from api.models.Resource import Resource, ResourceDataTable
from api.utils.data_indexing import index_resource_data

logger = structlog.get_logger("dataspace.commands.update_dataindex")


class Command(BaseCommand):
    help = "Update dataindex (ResourceDataTable) for all resources"

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "--resource-id",
            type=str,
            help="Update dataindex for a specific resource by ID",
            required=False,
            default="",
        )
        parser.add_argument(
            "--dataset-id",
            type=str,
            help="Update dataindex for all resources in a specific dataset by ID",
            required=False,
            default="",
        )
        parser.add_argument(
            "--skip-existing",
            action="store_true",
            help="Skip resources that already have a dataindex",
        )

    def handle(self, *args: Any, **options: Dict[str, Any]) -> None:
        resource_id_str: str = str(options.get("resource_id", ""))
        dataset_id_str: str = str(options.get("dataset_id", ""))
        skip_existing = options.get("skip_existing")

        # Convert string IDs to UUID objects if provided
        resource_id: Optional[uuid.UUID] = None
        dataset_id: Optional[uuid.UUID] = None

        # Only try to convert valid UUID strings
        if resource_id_str and resource_id_str.strip():
            try:
                resource_id = uuid.UUID(resource_id_str)
            except ValueError:
                self.stdout.write(
                    self.style.ERROR(
                        f"Invalid UUID format for resource-id: {resource_id_str}"
                    )
                )
                return

        if dataset_id_str and dataset_id_str.strip():
            try:
                dataset_id = uuid.UUID(dataset_id_str)
            except ValueError:
                self.stdout.write(
                    self.style.ERROR(
                        f"Invalid UUID format for dataset-id: {dataset_id_str}"
                    )
                )
                return

        start_time = time.time()
        success_count = 0
        error_count = 0
        skipped_count = 0

        # Filter resources based on command arguments
        resources = Resource.objects.all()

        if resource_id:
            resources = resources.filter(id=resource_id)
            self.stdout.write(f"Processing single resource with ID: {resource_id}")
        elif dataset_id:
            resources = resources.filter(dataset_id=dataset_id)
            self.stdout.write(f"Processing resources for dataset with ID: {dataset_id}")
        else:
            self.stdout.write("Processing all resources")

        total_resources = resources.count()
        self.stdout.write(f"Found {total_resources} resources to process")

        # Process each resource
        for i, resource in enumerate(resources, 1):
            # Cast resource to the proper type for type checking
            resource = cast(Resource, resource)

            try:
                # Skip resources that don't have file details
                if not hasattr(resource, "resourcefiledetails"):
                    self.stdout.write(
                        self.style.WARNING(
                            f"[{i}/{total_resources}] Skipping resource {resource.id} - No file details"
                        )
                    )
                    skipped_count += 1
                    continue

                # Skip resources that aren't CSV files
                file_details = resource.resourcefiledetails
                if not file_details or not file_details.format.lower() == "csv":
                    self.stdout.write(
                        self.style.WARNING(
                            f"[{i}/{total_resources}] Skipping resource {resource.id} - Not a CSV file"
                        )
                    )
                    skipped_count += 1
                    continue

                # Skip resources that already have a dataindex if skip_existing is True
                if skip_existing and hasattr(resource, "resourcedatatable"):
                    self.stdout.write(
                        self.style.WARNING(
                            f"[{i}/{total_resources}] Skipping resource {resource.id} - Already has dataindex"
                        )
                    )
                    skipped_count += 1
                    continue

                self.stdout.write(
                    f"[{i}/{total_resources}] Processing resource {resource.id} - {resource.name}"
                )

                # Update dataindex for the resource
                with transaction.atomic():
                    data_table = index_resource_data(resource)

                if data_table:
                    success_count += 1
                    self.stdout.write(
                        self.style.SUCCESS(
                            f"[{i}/{total_resources}] Successfully updated dataindex for resource {resource.id}"
                        )
                    )
                else:
                    error_count += 1
                    self.stdout.write(
                        self.style.ERROR(
                            f"[{i}/{total_resources}] Failed to update dataindex for resource {resource.id}"
                        )
                    )
            except Exception as e:
                error_count += 1
                logger.error(
                    f"Error updating dataindex for resource {resource.id}",
                    error=str(e),
                    exc_info=True,
                )
                self.stdout.write(
                    self.style.ERROR(
                        f"[{i}/{total_resources}] Error updating dataindex for resource {resource.id}: {str(e)}"
                    )
                )

        # Print summary
        elapsed_time = time.time() - start_time
        self.stdout.write(
            self.style.SUCCESS(
                f"\nDataindex update completed in {elapsed_time:.2f} seconds"
            )
        )
        self.stdout.write(f"Total resources processed: {total_resources}")
        self.stdout.write(f"Successful updates: {success_count}")
        self.stdout.write(f"Failed updates: {error_count}")
        self.stdout.write(f"Skipped resources: {skipped_count}")



from typing import List, Optional

import strawberry
import strawberry_django
from strawberry import Info, auto, field
from strawberry.enum import EnumType

from api.activities import collaborative
from api.models import (
    Collaborative,
    CollaborativeMetadata,
    CollaborativeOrganizationRelationship,
    Organization,
)
from api.types.base_type import BaseType
from api.types.type_collaborative_metadata import TypeCollaborativeMetadata
from api.types.type_collaborative_organization import (
    TypeCollaborativeOrganizationRelationship,
)
from api.types.type_dataset import TypeDataset, TypeTag
from api.types.type_geo import TypeGeo
from api.types.type_organization import TypeOrganization
from api.types.type_sdg import TypeSDG
from api.types.type_sector import TypeSector
from api.types.type_usecase import TypeUseCase
from api.utils.enums import CollaborativeStatus, OrganizationRelationshipType
from authorization.types import TypeUser

collaborative_status = strawberry.enum(CollaborativeStatus)  # type: ignore


@strawberry_django.filter(Collaborative)
class CollaborativeFilter:
    """Filter class for Collaborative model."""

    id: auto
    slug: auto
    status: Optional[collaborative_status]


@strawberry_django.order(Collaborative)
class CollaborativeOrder:
    """Order class for Collaborative model."""

    title: auto
    created: auto
    modified: auto


@strawberry_django.type(
    Collaborative,
    pagination=True,
    fields="__all__",
    filters=CollaborativeFilter,
    order=CollaborativeOrder,  # type:ignore
)
class TypeCollaborative(BaseType):
    """GraphQL type for Collaborative model."""

    user: TypeUser = strawberry.field(description="User who created this collaborative")
    organization: Optional[TypeOrganization] = strawberry.field(
        description="Organization associated with this collaborative"
    )
    platform_url: Optional[str] = strawberry.field(
        description="URL of the platform where this collaborative is published"
    )

    @strawberry.field(
        description="Check if this collaborative is created by an individual user."
    )
    def is_individual_collaborative(self) -> bool:
        """Check if this collaborative is created by an individual user."""
        return self.organization is None

    @strawberry.field(description="Get geographies associated with this collaborative.")
    def geographies(self) -> Optional[List[TypeGeo]]:
        """Get geographies associated with this collaborative."""
        try:
            queryset = self.geographies.all().order_by("name")  # type: ignore
            if not queryset.exists():
                return []
            return TypeGeo.from_django_list(queryset)
        except Exception:
            return []

    @strawberry.field(description="Get datasets associated with this collaborative.")
    def datasets(self) -> Optional[List["TypeDataset"]]:
        """Get datasets associated with this collaborative."""
        try:
            # Return raw Django objects and let Strawberry handle conversion
            queryset = self.datasets.all().order_by("-modified")  # type: ignore
            if not queryset.exists():
                return []
            return TypeDataset.from_django_list(queryset)
        except Exception:
            return []

    @strawberry.field(description="Get use cases associated with this collaborative.")
    def use_cases(self) -> Optional[List["TypeUseCase"]]:
        """Get use cases associated with this collaborative."""
        try:
            queryset = self.use_cases.all().order_by("-modified")  # type: ignore
            if not queryset.exists():
                return []
            return TypeUseCase.from_django_list(queryset)
        except Exception:
            return []

    @strawberry.field(
        description="Get the count of datasets associated with this collaborative."
    )
    def dataset_count(self: "TypeCollaborative", info: Info) -> int:
        """Get the count of datasets associated with this collaborative."""
        try:
            return self.datasets.count()  # type: ignore
        except Exception:
            return 0

    @strawberry.field(
        description="Get the count of use cases associated with this collaborative."
    )
    def use_case_count(self: "TypeCollaborative", info: Info) -> int:
        """Get the count of use cases associated with this collaborative."""
        try:
            return self.use_cases.count()  # type: ignore
        except Exception:
            return 0

    @strawberry.field(description="Get publishers associated with this collaborative.")
    def publishers(self) -> Optional[List["TypeOrganization"]]:
        """Get publishers associated with this collaborative."""
        try:
            queryset = Organization.objects.filter(datasets__in=self.datasets.all()).distinct()  # type: ignore
            if not queryset.exists():
                return []
            return TypeOrganization.from_django_list(queryset)
        except Exception:
            return []

    @strawberry.field(description="Get sectors associated with this collaborative.")
    def sectors(self) -> Optional[List["TypeSector"]]:
        """Get sectors associated with this collaborative."""
        try:
            queryset = self.sectors.all().order_by("name")  # type: ignore
            if not queryset.exists():
                return []
            return TypeSector.from_django_list(queryset)
        except Exception:
            return []

    @strawberry.field(description="Get SDGs associated with this collaborative.")
    def sdgs(self) -> Optional[List["TypeSDG"]]:
        """Get SDGs associated with this collaborative."""
        try:
            queryset = self.sdgs.all()  # type: ignore
            if not queryset.exists():
                return []
            return TypeSDG.from_django_list(queryset)
        except Exception:
            return []

    @strawberry.field(description="Get tags associated with this collaborative.")
    def tags(self) -> Optional[List["TypeTag"]]:
        """Get tags associated with this collaborative."""
        try:
            queryset = self.tags.all()  # type: ignore
            if not queryset.exists():
                return []
            return TypeTag.from_django_list(queryset)
        except Exception:
            return []

    @strawberry.field(description="Get metadata associated with this collaborative.")
    def metadata(self) -> Optional[List["TypeCollaborativeMetadata"]]:
        """Get metadata associated with this collaborative."""
        try:
            queryset = CollaborativeMetadata.objects.filter(collaborative=self)  # type: ignore
            if not queryset.exists():
                return []
            return TypeCollaborativeMetadata.from_django_list(queryset)
        except Exception:
            return []

    @strawberry.field(
        description="Get contributors associated with this collaborative."
    )
    def contributors(self) -> Optional[List["TypeUser"]]:
        """Get contributors associated with this collaborative."""
        try:
            queryset = self.contributors.all().order_by("first_name")  # type: ignore
            if not queryset.exists():
                return []
            return TypeUser.from_django_list(queryset)
        except Exception:
            return []

    @strawberry.field(
        description="Get organization relationships associated with this collaborative."
    )
    def organization_relationships(
        self,
    ) -> Optional[List["TypeCollaborativeOrganizationRelationship"]]:
        """Get organization relationships associated with this collaborative."""
        try:
            queryset = CollaborativeOrganizationRelationship.objects.filter(collaborative=self)  # type: ignore
            if not queryset.exists():
                return []
            return TypeCollaborativeOrganizationRelationship.from_django_list(queryset)
        except Exception:
            return []

    @strawberry.field(
        description="Get supporting organizations for this collaborative."
    )
    def supporting_organizations(self) -> Optional[List["TypeOrganization"]]:
        """Get supporting organizations for this collaborative."""
        try:
            relationships = (
                CollaborativeOrganizationRelationship.objects.filter(
                    collaborative=self,  # type: ignore
                    relationship_type=OrganizationRelationshipType.SUPPORTER,
                )
                .select_related("organization")
                .order_by("organization__name")
            )

            if not relationships.exists():
                return []

            organizations = [rel.organization for rel in relationships]  # type: ignore
            return TypeOrganization.from_django_list(organizations)
        except Exception:
            return []

    @strawberry.field(description="Get partner organizations for this collaborative.")
    def partner_organizations(self) -> Optional[List["TypeOrganization"]]:
        """Get partner organizations for this collaborative."""
        try:
            relationships = (
                CollaborativeOrganizationRelationship.objects.filter(
                    collaborative=self,  # type: ignore
                    relationship_type=OrganizationRelationshipType.PARTNER,
                )
                .select_related("organization")
                .order_by("organization__name")
            )

            if not relationships.exists():
                return []

            organizations = [rel.organization for rel in relationships]  # type: ignore
            return TypeOrganization.from_django_list(organizations)
        except Exception:
            return []
